from starling_sim.basemodel.agent.vehicles.service_vehicle import ServiceVehicle
from starling_sim.basemodel.trace.events import StaffOperationEvent


class RepositioningVehicle(ServiceVehicle):
    def process_stop_(self, stop):
        if stop.type == stop.REPOSITIONING:
            yield self.execute_process(self.process_repositioning_(stop))
            self.planning.remove(stop)
        elif stop.type is None:
            self.planning.remove(stop)
            return
        else:
            self.log_message("Unsupported stop type to process {}".format(stop.type), 30)

        # compute the dwell time duration
        dwell_time = self.compute_dwell_time(stop)

        yield self.execute_process(self.spend_time_(dwell_time))

    # TODO : refactor and cleanup
    def process_repositioning_(self, stop):
        """
        Realise the repositioning operation described by the stop.

        If 'total' attribute is negative, the service vehicle takes
        objects from the station. The operations are realised respecting
        the service vehicle and station capacity constraints.

        :param stop: stop with type REPOSITIONING
        """

        if stop.station not in self.operator.stations:
            return

        station = self.operator.stations[stop.station]
        self.log_message("Processing stop {}".format(stop))
        successful_operations = 0
        targets = []
        if stop.total < 0:
            for i in range(-stop.total):
                if self.load() == self.seats:
                    break

                request = station.get_from_store(self)

                # TODO : works with 0 ?
                result = yield request.event_ | self.sim.scheduler.timeout(0)

                if request.event_ in result:
                    request.success = True
                    request.result = result[request.event_]
                    # end request
                    request.event_.cancel()

                    successful_operations -= 1
                    targets.append(request.result)
                    self.add_passenger(request.result)
                else:
                    break

        else:
            for i in range(stop.total):
                if self.load() == 0:
                    break

                occupant = self.occupants[-1]
                request = station.return_to_store(self, occupant)

                # TODO : works with 0 ?
                result = yield request.event_ | self.sim.scheduler.timeout(0)

                if request.event_ in result:
                    request.success = True
                    # end request
                    request.event_.cancel()

                    successful_operations += 1
                    targets.append(occupant)
                    self.remove_passenger(occupant)
                else:
                    break

        operation_event = StaffOperationEvent(
            self.sim.scheduler.now(),
            self,
            successful_operations,
            stop.total,
            targets=targets,
            structure=station,
        )
        self.trace_event(operation_event)

    # def repositioning(self, action_type, station):
    #
    #     # get the action information and request
    #     if action_type == -1:
    #         limit = self.seats
    #         request = station.get_from_store(self)
    #     elif action_type == 1:
    #         limit = 0
    #         occupant = self.occupants[-1]
    #         request = station.return_to_store(self, occupant)
    #     else:
    #         self.log_message("Unknown action type for repositioning {}".format(action_type), 30)
    #         return False
    #
    #     # try to execute the repositioning action
    #     result = yield request.event | self.sim.scheduler.timeout(0)
    #
    #     if request.event in result:
    #         request.success = True
    #         request.result = result[request.event]
    #         # end request
    #         request.event.cancel()
    #         self.get_passenger(request.result)
    #         if
    #
    #     else:
    #         break



"""
This module contains utils for the Starling framework.
"""

import os
import subprocess
import logging
import json
import geopandas
import pandas as pd
import numpy as np
import gtfs_kit as gt
import osmnx as ox
import gzip
import shutil
import copy
from shapely.geometry import Polygon, LineString
from numbers import Integral
from jsonschema import validate, ValidationError
from starling_sim.utils.paths import (
    schemas_folder,
    gtfs_feeds_folder,
    osm_graphs_folder,
    scenario_inputs_folder,
    PARAMETERS_FILENAME,
    INPUT_FOLDER_NAME,
)
from starling_sim.utils.simulation_logging import BLANK_LOGGER

pd.set_option("display.expand_frame_repr", False)


# Starling exceptions


class StarlingException(Exception):
    """
    Base class for all Starling exceptions.
    """


class LeavingSimulation(StarlingException):
    """
    Exception raised by agents for leaving the simulation.

    Agents should raise this exception by calling their leave_simulation() method
    to leave their loop and terminate their SimPy process.

    LeavingSimulation exceptions are caught in the agent :meth:`~starling_sim.basemodel.agent.agent.Agent.simpy_loop_`
    method so a LeavingSimulationEvent can be traced and the agent main process can terminate.
    """


class SimulationError(LeavingSimulation):
    """
    Simulation error exception.

    This exception should be used when an unwanted
    event occurs in a simulation, like saying "We shouldn't be here".
    """


class PlanningChange(Exception):
    """
    Exception raised to interrupt a service vehicle loop
    when its planning has changed.
    """

    DEFAULT_MESSAGE = (
        "The service vehicle planning has been changed during "
        "service activity. Catch this exception to manage this case."
    )

    def __init__(self, msg=DEFAULT_MESSAGE, *args):
        super().__init__(msg, *args)


# json utils


def json_dump(data, filepath):
    """
    Create a json file and dump the given object

    :param data:
    :param filepath:
    :return:
    """

    with open(filepath, "w") as outfile:
        json.dump(data, outfile)


def json_pretty_dump(data, filepath):
    """
    Creates a json file and dumps the given object
    in a pretty print way

    :param data: object to be dumped into the json file
    :param filepath: path to the json file
    """

    with open(filepath, "w") as outfile:
        outfile.write(json.dumps(data, indent=4))


# TODO : check that the file contains a json, or at least catch exception
def json_load(filepath):
    """
    Loads the content of the given json file

    :param filepath: path to the json file
    :return: object loaded from json file
    """

    with open(filepath, "r") as param_file:
        return json.load(param_file)


# compression utils


def gz_compression(filepath, delete_source=True):
    """
    Compress the given file using gzip.

    Delete the source file if asked.

    :param filepath: path pointing to the source file
    :param delete_source: boolean indicating if the source file should be deleted
    """

    # compress the file using gzip
    compressed_path = filepath + ".gz"
    with open(filepath, "rb") as f_in:
        with gzip.open(compressed_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    # delete source file if asked
    if delete_source:
        os.remove(filepath)

    return compressed_path


def gz_decompression(filepath, delete_source=True):
    """
    Decompress the given file using gzip.

    Delete the source file if asked.

    :param filepath: path pointing to the source file
    :param delete_source: boolean indicating if the source file should be deleted

    :raises: ValueError if the filepath does not end with '.gz'
    """

    # test if the file ends with .gz
    if not filepath.endswith(".gz"):
        logging.log(
            30,
            "File to decompress does not end with '.gz'. Continuing without decompressing the file.",
        )

    # decompress the file unsing gzip
    with gzip.open(filepath, "rb") as f_in:
        with open(filepath[:-3], "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    # delete source file if asked
    if delete_source:
        os.remove(filepath)


# starling file information


def create_file_information(
    filepath,
    mimetype,
    compressed_mimetype=None,
    content=None,
    subject=None,
    keep_filepath=False,
):
    """
    Create a dict containing file information.

    :param filepath: file path
    :param mimetype: file mimetype
    :param compressed_mimetype: file mimetype after decompression
    :param content: content metadata
    :param subject: subject metadata
    :param keep_filepath: keep complete file path in 'filename'

    :return: { filename/filepath, mimetype, metadata }
    """

    if content is None:
        raise ValueError("'content' metadata was not provided for file {}".format(filepath))

    if compressed_mimetype is None:
        compressed_mimetype = mimetype

    metadata = {"compressed-mimetype": compressed_mimetype, "content": content}

    if subject is not None:
        metadata["subject"] = subject

    file_information = {
        "filename": filepath if keep_filepath else os.path.basename(filepath),
        "mimetype": mimetype,
        "metadata": metadata,
    }

    return file_information


# json schema validation

# use the type check from Draft4Validator, because Draft7 considers 1.0 as integer
# CustomValidator = validators.extend(Draft7Validator, type_checker=Draft4Validator.TYPE_CHECKER)


def validate_against_schema(instance, schema, raise_exception=True):
    # load schema object
    schema = load_schema(schema, False)

    # validate against schema and catch eventual exception
    try:
        validate(instance, schema)
        return True
    except ValidationError as e:
        if raise_exception:
            raise e
        else:
            logging.log(
                30,
                "JSON Schema validation of :\n\n{}\n\nfailed with message:\n\n {} ".format(
                    instance, e
                ),
            )
            return False


def add_defaults_and_validate(instance, schema, raise_exception=True):
    # initialise an empty result dict
    res = copy.deepcopy(instance)

    # load schema object
    schema = load_schema(schema)

    # add default properties to the schema
    add_defaults(res, schema)

    # validate the final instance against the schema
    validate_against_schema(res, schema, raise_exception)

    return res


def add_defaults(instance, schema, current_prop=None):
    if "type" in schema and schema["type"] == "object" and "default" not in schema:
        for prop in schema["properties"].keys():
            if current_prop is None:
                prop_instance = instance
            else:
                if current_prop in instance:
                    prop_instance = instance[current_prop]
                else:
                    prop_instance = dict()
                    instance[current_prop] = prop_instance

            add_defaults(prop_instance, schema["properties"][prop], prop)

    else:
        if current_prop not in instance and "default" in schema:
            instance[current_prop] = schema["default"]


def load_schema(schema, make_copy=True):
    if isinstance(schema, str):
        # load the schema if a path is provided
        final_schema = json_load(schemas_folder() + schema)
    elif isinstance(schema, dict):
        if make_copy:
            final_schema = copy.deepcopy(schema)
        else:
            final_schema = schema
    else:
        raise TypeError("The provided schema is neither a dict nor a path to a schema file")

    return final_schema


# converters


def get_sec(time_str):
    """Get Seconds from time."""

    # HH:MM:SS to seconds
    h, m, s = time_str.split(":")
    return int(h) * 3600 + int(m) * 60 + int(s)


def make_int(a):
    """
    Convert the given integer to the Python int type.

    Check that the given number is an Integral, and raise a
    ValueError if it is not the case.

    :param a: integer to check

    :return: integer with Python int type
    """

    if isinstance(a, float):
        if not a.is_integer():
            raise ValueError("Object {} is not Integral".format(a))
    elif not isinstance(a, Integral):
        raise ValueError("Object {} is not Integral, cannot convert to int".format(a))

    return int(a)


# geojson utils


def new_point_feature(point_localisation=None, properties=None):
    """

    :param point_localisation:
    :param properties:
    :return:
    """

    # set point localisation to [0, 0] by default
    if point_localisation is None:
        point_localisation = [0, 0]

    # create  the feature geometry
    geometry = {"type": "Point", "coordinates": point_localisation}

    return new_feature(geometry, properties)


def new_line_string_feature(linestring, properties=None):
    # create  the feature geometry
    geometry = {"type": "LineString", "coordinates": linestring}

    return new_feature(geometry, properties)


def new_multi_polygon_feature(polygon_list, properties=None):
    """


    :param polygon_list: list of polygon coordinates.
    :param properties:
    :return:
    """

    # create  the feature geometry
    geometry = {"type": "MultiPolygon", "coordinates": polygon_list}

    return new_feature(geometry, properties)


def new_feature(geometry, properties=None):
    # set properties with empty dict by default
    if properties is None:
        properties = dict()

    # build a geojson feature
    feature = {"type": "Feature", "geometry": geometry, "properties": properties}

    return feature


def new_feature_collection(features=None):
    """

    :param features:
    :return:
    """

    # set features to empty list by default
    if features is None:
        features = []

    # build a geojson feature collection
    feature_collection = {"type": "FeatureCollection", "features": features}

    return feature_collection


# shapes functions


def shapely_polygon_from_points(points, reverse=False):
    """
    Create and return a Shapely Polygon object.

    The polygon is initialised using the given list of tuples
    representing coordinates. If reverse is True, the coordinates
    values are inverted. This can be used to set GPS coordinates in
    order, for instance.

    :param points: list of lists containing point coordinates
    :param reverse: boolean indicating if coordinates should be reversed.
        Default is False.

    :return: Shapely Polygon object
    """

    # if asked, transform (lat, lon) points into (lon, lat) ones
    if reverse:
        points = [point[::-1] for point in points]

    # create a shapely Polygon from point list
    polygon = Polygon(points)

    return polygon


def geopandas_points_from_localisations(localisations):
    """
    Create and return a GeoDataFrame containing shapely Point objects.

    localisations can either be a tuple or list describing a single
    (lat, lon) point, or a Dataframe containing a 'lat' and a 'lon' column.

    The localisations must be specified using "epsg:4326"

    :param localisations: either a single point (lat, lon) coordinates,
        or a Dataframe containing a 'lat' and a 'lon' column, "epsg:4326"

    :return: GeoDataFrame containing shapely Point objects with crs="epsg:4326"
    """

    if isinstance(localisations, tuple) or isinstance(localisations, list):
        # create a DataFrame containing the point coordinates
        localisations = pd.DataFrame(
            {"lat": localisations[0], "lon": localisations[1]}, columns=["lat", "lon"], index=[0]
        )

    # create a GeoDataFrame from df
    gdf = geopandas.GeoDataFrame(
        localisations, geometry=geopandas.points_from_xy(localisations.lon, localisations.lat)
    )

    # set epsg
    gdf.set_crs("epsg:4326")

    return gdf


def geopandas_polygon_from_points(points):
    """
    Create and return a GeoDataFrame containing a single polygon.

    The GeoDataFrame contains a Shapely Polygon object created using
    the given points, and stored in the first line of the 'geometry' column.

    The point coordinates are reversed to have a zone specified with
    (lon, lat) points, with "epsg:4326"

    :param points: list of coordinates, must be (lon, lat) points.

    :return: GeoDataFrame containing a shapely Polygon with crs="epsg:4326"
    """

    # create a shapely Polygon from point list
    polygon = shapely_polygon_from_points(points)

    # create a GeoDataFrame containing the polygon
    gdf = geopandas.GeoDataFrame([polygon], columns=["geometry"])

    # set epsg, not in GeoDataFrame init
    gdf.set_crs("epsg:4326")

    return gdf


def points_in_zone(localisations, zone):
    """
    Evaluate if the given points are contained in the zone.

    localisations can either be a tuple or list describing a single
    (lat, lon) point, or a DataFrame containing a 'lat' and a 'lon' column.

    The function either returns a boolean if a single point is provided,
    or adds a "in_zone" column to the DataFrame containing booleans.

    :param localisations: either a single point (lat, lon) coordinates,
        or a DataFrame containing a 'lat' and a 'lon' column, "epsg:4326"
    :param zone: GeoDataFrame containing a shapely Polygon with crs="epsg:4326"

    :return: either a boolean, or the localisations DataFrame with an
        additional "in_zone" column containing booleans.
    """

    # create a GeoDataFrame point
    geopandas_points = geopandas_points_from_localisations(localisations)

    # set epsg
    geopandas_points.set_crs("epsg:4326")

    # evaluate if within zone
    res = geopandas.sjoin(geopandas_points, zone, how="left", predicate="within")

    # set new column 'in_zone'
    res["in_zone"] = res["index_right"].replace(to_replace={np.nan: False, 0: True})

    # remove evaluation columns
    res.drop(labels=["geometry", "index_right"], axis=1, inplace=True)

    # if a single point was provided, return a the boolean
    if isinstance(localisations, tuple) or isinstance(localisations, list):
        res = res.loc[res.index[0], "in_zone"]

    return res


def stops_table_from_geojson(geojson_path):
    """
    Create a stops table from the provided geojson file.

    :param geojson_path:

    :return: a DataFrame containing the stops information
    """

    # get the feature collection from the geojson
    stops_feature_collection = json_load(geojson_path)

    # initialise the table
    stops_table = pd.DataFrame()

    for i in range(len(stops_feature_collection["features"])):
        feature = stops_feature_collection["features"][i]
        properties = feature["properties"]

        # get coordinates from the geometry
        properties["stop_lon"] = feature["geometry"]["coordinates"][0]
        properties["stop_lat"] = feature["geometry"]["coordinates"][1]

        # create a stop id if not provided
        if "stop_id" not in properties:
            properties["stop_id"] = str(i)

        # create a stop name if not provided
        if "stop_name" not in properties:
            properties["stop_name"] = "Stop point {id}".format(id=properties["stop_id"])

        # convert properties to DataFrame
        properties = pd.DataFrame([properties])

        # append a new row
        stops_table = pd.concat([stops_table, properties], ignore_index=True)

    return stops_table


def stop_table_from_gtfs(
    gtfs_feed, routes=None, zone=None, fixed_stops=None, active_stops_only=False
):
    result_table = pd.DataFrame()

    gtfs_stops = gtfs_feed.get_stops().copy()

    if fixed_stops is not None:
        result_table = pd.concat(
            [result_table, gtfs_stops[gtfs_stops["stop_id"].isin(fixed_stops)]], sort=False
        )

    if active_stops_only:
        stop_times = gtfs_feed.get_stop_times()
        stop_ids = stop_times.drop_duplicates("stop_id")
        gtfs_stops = pd.merge(gtfs_stops, stop_ids)[gtfs_stops.columns]

    if routes is not None:
        gtfs_stops = gtfs_feed.get_stops(route_ids=routes)

    if zone is not None:
        gtfs_stops.rename(columns={"stop_lat": "lat", "stop_lon": "lon"}, inplace=True)
        gtfs_stops = points_in_zone(gtfs_stops, zone)
        gtfs_stops.rename(columns={"lat": "stop_lat", "lon": "stop_lon"}, inplace=True)

        gtfs_stops = gtfs_stops[gtfs_stops["in_zone"]]

        # remove stops areas
        gtfs_stops = gtfs_stops[
            (gtfs_stops["location_type"] == 0) | (gtfs_stops["location_type"].isna())
        ]

    result_table = pd.concat([result_table, gtfs_stops], sort=False)
    result_table.drop_duplicates(inplace=True)

    return result_table


def import_osm_graph(
    method,
    network_type,
    simplify,
    query=None,
    which_result=None,
    point=None,
    dist=None,
    polygon=None,
    outfile=None,
):
    """
    Generate an OSM graph from given parameters and store it in a file.

    The osmnx function used to generate the OSM graph is specified by the method parameter.

    The correct parameters must be provided according to the import method.

    :param method: name of the osmnx import method. Among ['place', 'point', 'polygon'].
    :param network_type: OSM network_type of the graph
    :param simplify: boolean indicating if the graph should be simplified
    :param query: string, dict or list describing the place (must be geocodable)
    :param which_result: integer describing which geocoding result to use,
        or None to auto-select the first (Multi)Polygon
    :param point: [lon, lat] coordinates of the center point
    :param dist: distance of the bbox from the center point
    :param polygon: list of points describing a polygon
    :param outfile: optional name for the output file
    """

    # import the OSM graph according to the given method
    if method == "place":
        graph = osm_graph_from_place(query, which_result, network_type, simplify)
        default_outfile = "G{}_{}.graphml.bz2".format(network_type, query)

    elif method == "point":
        graph = osm_graph_from_point(point, dist, network_type, simplify)
        default_outfile = "G{}_{}-{}_{}.graphml.bz2".format(network_type, point[0], point[1], dist)

    elif method == "polygon":
        default_outfile = None
        if outfile is None:
            print("Outfile name must be provided when importing from polygon.")
            return
        graph = osm_graph_from_polygon(polygon, network_type, simplify)

    else:
        print("Unknown import method {}. Choices are ['point', 'place', 'polygon'].")
        return

    # keep the largest strongly connected component of the graph
    graph = ox.utils_graph.get_largest_component(graph, strongly=True)

    # get the output filename
    if outfile is None:
        # add 'S' to simplified graphs
        if simplify:
            default_outfile = "S" + default_outfile

        outfile = default_outfile

    # save the graph at .graphml format
    save_osm_graph(graph, filename=outfile, folder=osm_graphs_folder())

    # return the graph
    return graph


def osm_graph_from_point(point, distance, network_type, simplify):
    """
    Import an OSM graph of an area around the location point.

    The import is done with the distance_type parameter set to 'bbox'.

    :param point: (lon, lat) point
    :param distance: distance around point
    :param network_type: osm network type
    :param simplify: boolean indicating if the graph should be simplified

    :return: a networkx graph
    """

    if point is None or distance is None:
        print(
            "The point and distance parameters must be specified when importing graph from point."
        )
        exit(1)

    # reverse the point coordinates (osmnx takes (lat, lon) coordinates)
    point = (point[1], point[0])

    return ox.graph_from_point(
        point, dist=distance, dist_type="bbox", network_type=network_type, simplify=simplify
    )


def osm_graph_from_polygon(polygon_points, network_type, simplify):
    """
    Import an OSM graph of the area within the polygon.

    :param polygon_points: list of (lon, lat) points delimiting the network zone
    :param network_type: osm network type
    :param simplify: boolean indicating if the graph should be simplified

    :return: a networkx graph
    """

    if polygon_points is None:
        print("The polygon parameter must be specified when importing graph from polygon.")
        exit(1)

    # create a shapely polygon with (lat, lon) coordinates from the list of points
    shapely_polygon = shapely_polygon_from_points(polygon_points)

    return ox.graph_from_polygon(shapely_polygon, network_type=network_type, simplify=simplify)


def osm_graph_from_place(query, which_result, network_type, simplify):
    """
    Import an OSM graph of the area described by the geocodable query.

    :param query: string, dict or list describing the place (must be geocodable)
    :param which_result: integer describing which geocoding result to use,
        or None to auto-select the first (Multi)Polygon
    :param network_type: osm network type
    :param simplify: boolean indicating if the graph should be simplified

    :return: a networkx graph
    """

    if query is None:
        print("The query parameter must be specified when importing graph from place.")
        exit(1)

    return ox.graph_from_place(
        query, network_type=network_type, simplify=simplify, which_result=which_result
    )


def save_osm_graph(graph, filename, folder):
    """
    Save the given graph in a .graphml file.

    Detect if the filename ends with '.bz2', and realise
    a bz2 compression accordingly.

    :param graph: saved graph
    :param filename: name of the save file
    :param folder: name of the save folder
    """

    # check bz2 extension
    if filename.endswith(".bz2"):
        to_bz2 = True
        filename = filename[:-4]
    else:
        to_bz2 = False

    # check filename
    if not filename.endswith(".graphml"):
        raise ValueError("OSM graph filename must end with .graphml or .graphml.bz2")

    # save the graph
    filepath = folder + filename
    ox.save_graphml(graph, filepath)

    # compress to bz2 if necessary
    if to_bz2:
        subprocess.run(["bzip2", "-z", "-f", filepath])
        print("Saved osm graph at " + filepath + ".bz2")
    else:
        print("Saved osm graph at " + filepath)


def osm_graph_from_file(filename, folder=None):
    """
    Import an osm graph from a .graphml file.

    :param filename: path to the .graphml file
    :param folder: folder containing the .graphml file.
        Default is the osm graphs folder.

    :return: a networkx graph, or None if import fails
    """

    if folder is None:
        folder = osm_graphs_folder()

    filepath = folder + filename

    return ox.load_graphml(filepath)


# gtfs utils


def import_gtfs_feed(gtfs_filename, transfer_restriction=None, folder=None):
    """
    Import a gtfs feed from the given file.

    Also check that stop times are ordered and transfers are symmetrical.

    :param gtfs_filename: name of a gtfs file
    :param transfer_restriction: duration restriction on the transfers
    :param folder: folder where the gtfs is stored.
        Default is the GTFS_FEEDS_FOLDER.

    :return: gtfs-kit Feed object
    """

    # read the gtfs feed using gtfs-kit
    if folder is None:
        folder = gtfs_feeds_folder()
    path = folder + gtfs_filename
    feed = gt.read_feed(path, dist_units="km")

    # additional operations and validations

    # order stop_times by trip_id and stop_sequence
    stop_times = feed.stop_times
    if stop_times is not None:
        stop_times = stop_times.sort_values(by=["trip_id", "stop_sequence"])
        feed.stop_times = stop_times
        feed.stop_times = feed.stop_times

    # check that foot-path transfers are symmetrical
    if feed.transfers is not None:
        transfer_table = feed.transfers.copy()
        transfer_table = transfer_table[
            transfer_table["from_stop_id"] != transfer_table["to_stop_id"]
        ]
        transfer_table["stop_A"] = transfer_table[["from_stop_id", "to_stop_id"]].apply(min, axis=1)
        transfer_table["stop_B"] = transfer_table[["from_stop_id", "to_stop_id"]].apply(max, axis=1)
        count = transfer_table.groupby(["stop_A", "stop_B"], as_index=False).agg(["count"])
        counts_not_equal_to_2 = count[count["min_transfer_time"]["count"] != 2]
        if not counts_not_equal_to_2.empty:
            logging.warning(
                "Transfer table of {} is not symmetrical (in term of arcs, not transfer times)".format(
                    gtfs_filename
                )
            )
        if not is_transitive(feed.transfers) and transfer_restriction is not None:
            feed.transfers = transitively_closed_transfers(feed.transfers, transfer_restriction)
    else:
        logging.warning("The given GTFS has no transfer table")

    return feed


def transitively_closed_transfers(transfers, restrict_transfer_time):
    """
    Restrict the transfer table under the given time limit and then make it transitive.

    The restriction on the transfer times help keeping the transfer table at a reasonable size
    after making it transitive.

    :param transfers: transfer table
    :param restrict_transfer_time: transfer duration limit

    :return: transitive closure of the restricted transfer table
    """

    # restrict original transfers so the final set of transfers isn't too large
    if restrict_transfer_time is not None:
        logging.info(
            "The GTFS transfer table is not transitively closed. "
            "Transfers are restricted to duration under {} seconds and then made transitive.".format(
                restrict_transfer_time
            )
        )
        transfers = transfers[transfers["min_transfer_time"] <= restrict_transfer_time]
        return make_transfers_transitively_closed(transfers)
    # else:
    #     new_transfers = None
    #     restrict_transfer_time = 60
    #     computations_too_big = False
    #     while not computations_too_big:
    #         restricted_transfers = transfers[transfers["min_transfer_time"] <= restrict_transfer_time]
    #         try:
    #             new_transfers = make_transfers_transitively_closed(restricted_transfers)
    #             restrict_transfer_time += 60
    #         except TimeoutError as e:
    #             computations_too_big = True
    #             if restrict_transfer_time == 60:
    #                 raise e
    #
    #     return new_transfers


def is_transitive(transfers):
    """
    Test if the given transfer table is transitive.

    :param transfers: transfer table

    :return: boolean indicating if the table is transitive
    """

    if transfers is None:
        return True

    nb_transfers = len(transfers)
    # create new transfers using transitive property
    new_transfers = pd.merge(transfers, transfers, left_on="to_stop_id", right_on="from_stop_id")
    new_transfers["min_transfer_time"] = (
        new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
    )
    new_transfers["from_stop_id"] = new_transfers["from_stop_id_x"]
    new_transfers["to_stop_id"] = new_transfers["to_stop_id_y"]
    if "transfer_type" in transfers.columns:
        new_transfers["transfer_type"] = new_transfers[["transfer_type_x", "transfer_type_y"]].max(
            axis=1
        )
    new_transfers = new_transfers[transfers.columns]

    # add them to the former set of transfers
    transfers = pd.concat([transfers, new_transfers])

    # keep the best transfer for each couple of stops
    transfers.sort_values(by="min_transfer_time", inplace=True)
    transfers = transfers.groupby(by=["from_stop_id", "to_stop_id"], as_index=False).first()

    # break condition
    return len(transfers) == nb_transfers


def make_transfers_transitively_closed(transfers):
    """
    Make the given transfer table transitive by incrementally adding transition transfers.

    If transfers from A to B and from B to C exist, create a transfer from A to C.
    If several transfers from A to C exist, keep the one with shorter transfer duration.
    Repeat.

    :param transfers: transfer table

    :return: transitive closure of the given transfer table
    """

    # loop until no new transfer is found
    nb_transfers = len(transfers)
    while True:
        # create new transfers using transitive property
        new_transfers = pd.merge(
            transfers, transfers, left_on="to_stop_id", right_on="from_stop_id"
        )
        new_transfers["min_transfer_time"] = (
            new_transfers["min_transfer_time_x"] + new_transfers["min_transfer_time_y"]
        )
        new_transfers["from_stop_id"] = new_transfers["from_stop_id_x"]
        new_transfers["to_stop_id"] = new_transfers["to_stop_id_y"]
        if "transfer_type" in transfers.columns:
            new_transfers["transfer_type"] = new_transfers[
                ["transfer_type_x", "transfer_type_y"]
            ].max(axis=1)
        new_transfers = new_transfers[transfers.columns]

        # add them to the former set of transfers
        transfers = pd.concat([transfers, new_transfers])

        # keep the best transfer for each couple of stops
        transfers.sort_values(by="min_transfer_time", inplace=True)
        transfers = transfers.groupby(by=["from_stop_id", "to_stop_id"], as_index=False).first()

        # break condition
        if len(transfers) == nb_transfers:
            break

        nb_transfers = len(transfers)

    return transfers


def stops2geojson(stops_df):
    """
    Create a geojson FeatureCollection from the GTFS stops.

    :param stops_df: stops DataFrame of the GTFS

    :return: feature collection
    """

    features = []
    stops_df.apply(
        lambda x: features.append(
            new_point_feature(
                point_localisation=[x["stop_lon"], x["stop_lat"]],
                properties={
                    "stop_id": x["stop_id"],
                    "stop_name": x["stop_name"],
                },
            )
        ),
        axis=1,
    )

    return new_feature_collection(features)


def transfers2geojson(transfers_df, stops_df):
    """
    Create a geojson FeatureCollection from the GTFS transfers.

    :param transfers_df: transfers DataFrame of the GTFS
    :param stops_df: stops DataFrame of the GTFS

    :return: feature collection
    """

    # add the stop information to the from_stop and to_stop
    geo_transfers = pd.merge(transfers_df, stops_df, left_on=["from_stop_id"], right_on=["stop_id"])
    geo_transfers = pd.merge(geo_transfers, stops_df, left_on=["to_stop_id"], right_on=["stop_id"])
    geo_transfers = geo_transfers[
        [
            "from_stop_id",
            "to_stop_id",
            "min_transfer_time",
            "stop_lat_x",
            "stop_lon_x",
            "stop_lat_y",
            "stop_lon_y",
        ]
    ]

    # create the geometry attribute
    geo_transfers["geometry"] = geo_transfers.apply(
        lambda x: LineString(
            [(x["stop_lon_x"], x["stop_lat_x"]), (x["stop_lon_y"], x["stop_lat_y"])]
        ),
        axis=1,
    )
    geo_transfers = geopandas.GeoDataFrame(geo_transfers)
    geo_transfers = geo_transfers[["min_transfer_time", "geometry"]]

    geojson = geo_transfers.to_json(drop_id=True)

    return geojson


# folder creation


def create_if_not_exists(folder):
    """
    Test the existence of the folder and create it if not present.

    :param folder: path of the folder
    """

    if not os.path.exists(folder):
        os.mkdir(folder)
        return True
    else:
        return False


# console log


def display_horizontal_bar(lvl=20):
    """
    Display a horizontal bar in the terminal.

    Try to make it the size of the terminal width.

    :param lvl: log level
    """

    try:
        bar_size = os.get_terminal_size().columns
    except OSError:
        bar_size = 100

    BLANK_LOGGER.log(lvl, "\u2014" * bar_size)


# multiple scenarios


def create_sub_scenarios(simulation_scenario):
    logging.info("Creating sub scenarios of: " + simulation_scenario.name)

    nb_scenarios = simulation_scenario["multiple"]
    scenarios_folder = os.path.join(simulation_scenario.scenario_folder, "scenarios")
    create_if_not_exists(scenarios_folder)

    # set random seed
    np.random.seed(simulation_scenario["seed"])
    seeds = np.random.choice(range(100000), size=nb_scenarios, replace=False)
    sub_scenario_name_format = "{base_scenario}-{index}"

    scenario_paths = []
    for i in range(nb_scenarios):
        sub_scenario_name = sub_scenario_name_format.format(
            base_scenario=simulation_scenario.name, index=i
        )
        sub_scenario_folder = os.path.join(scenarios_folder, sub_scenario_name)
        sub_scenario_inputs_folder = scenario_inputs_folder(sub_scenario_folder)
        scenario_paths.append(sub_scenario_folder)

        if os.path.exists(sub_scenario_folder):
            logging.info("Scenario {} already created".format(sub_scenario_name))
            continue
        else:
            logging.info("Creating scenario " + sub_scenario_name)

        # create sub scenario folders
        create_if_not_exists(sub_scenario_folder)
        create_if_not_exists(sub_scenario_inputs_folder)

        # sub scenario inputs
        sub_parameters = simulation_scenario.copy_parameters()
        sub_parameters["scenario"] = sub_scenario_name
        sub_parameters["seed"] = int(seeds[i])
        del sub_parameters["multiple"]
        json_pretty_dump(
            sub_parameters, os.path.join(sub_scenario_inputs_folder, PARAMETERS_FILENAME)
        )

        # create symlinks to original inputs
        for input_file in os.listdir(simulation_scenario.inputs_folder):
            if input_file == PARAMETERS_FILENAME:
                continue
            input_filepath = os.path.join("..", "..", "..", INPUT_FOLDER_NAME, input_file)
            os.symlink(input_filepath, os.path.join(sub_scenario_inputs_folder, input_file))

    logging.info("End of sub scenarios creation\n")

    return scenario_paths



"""
This module is dedicated to discrete event scheduling.
It allows the representation of a time flow inside the simulation.
"""

import simpy
import logging
from numbers import Integral


# For now, only one class exists, using the SimPy_ library. If other scheduling classes should be developed,
# we should make them inherit from a super class containing the common methods.
class Scheduler:
    """
    This class manages a discrete event schedule using the SimPy_ library.

    Along with containing a SimPy environment, this class also provides methods
    for the generation of SimPy Event objects (Process objects are events too).
    Yielding these events will suspend the current activity (e.g agent loop), which will
    resume once the event occurs (or before if the process is interrupted).

    This way, agents can yield processes corresponding to an action (move, wait, etc)
    and resume to their loop after the action has finished.

    .. _SimPy: https://simpy.readthedocs.io/en/latest/index.html
    """

    def __init__(self):
        """
        Creation of a SimPy environment and a runtime limit (set to 0).
        """
        self.env = simpy.Environment()
        self.time_limit = 0

    def run(self, run_time):
        """
        Update the runtime limit attribute and run the environment.

        :param run_time: duration of the simulation run
        """

        self.time_limit += run_time
        self.env.run(run_time)

    def now(self):
        """
        Get the current simulation step.

        :return: Current time of the simulation (arbitrary unit)
        """

        return self.env.now

    def timeout(self, duration="inf"):
        """
        Create a SimPy timeout event with the given duration

        :param duration: time to pass, in simulation units. Defaults to 'inf',
         ie end at the simulation time limit. Must be positive and integer-like
        :return: SimPy timeout event
        """

        if duration == "inf":
            duration = self.time_limit - self.now()

        if not isinstance(duration, Integral) or duration < 0:
            logging.error("Incorrect waiting duration given to scheduler")
            raise ValueError(duration)

        return self.env.timeout(duration)

    def new_event_object(self):
        """
        Create and return a new Event object

        :return: Event object
        """

        return simpy.events.Event(self.env)

    def new_process(self, generator):
        """
        Create a SimPy process from the given generator

        :param generator: Python generator (function containing the yield keyword)
        :return: SimPy Process object
        """

        return self.env.process(generator)

    def interrupt(process, message=""):
        """
        Interrupt the given SimPy process, e.g. an agent waiting in a queue

        :param process: simPy Process object
        :param message: interruption message, to be displayed in the agent's trace
        """

        process.interrupt(message)

    interrupt = staticmethod(interrupt)



from starling_sim.basemodel.agent.agent import Agent


class SpatialAgent(Agent):
    """
    Class describing a spatial agent, with a position and origin in the simulation environment.
    """

    SCHEMA = {
        "properties": {
            "origin": {
                "type": ["number", "string"],
                "title": "Origin position",
                "description": "origin position id (inferred from geometry)",
            }
        },
        "required": ["origin"],
    }

    def __init__(self, simulation_model, agent_id, origin, **kwargs):
        Agent.__init__(self, simulation_model, agent_id, **kwargs)

        self.origin = origin

        self.position = origin

    def __str__(self):
        return "[id={}, origin={}]".format(self.id, self.origin)



from starling_sim.utils.testing import get_test_scenarios, SIMULATION_TEST_DATA_FOLDER
from starling_sim.utils import paths


## Model testing setup

# set data folder for all test execution
paths._DATA_FOLDER = SIMULATION_TEST_DATA_FOLDER


# add a 'models' option to pytest command line
# allows specifying model codes to test
def pytest_addoption(parser):
    parser.addoption(
        "--models",
        nargs="*",
        metavar=("MODEL_CODE_1", "MODEL_CODE_2"),
        default=None,
        help="list of model codes to run tests on",
    )


# parameterize test_model_scenario test function
def pytest_generate_tests(metafunc):
    if metafunc.function.__name__ == "test_model_scenario":
        # list test scenarios present in the provided models' folders
        scenarios = get_test_scenarios(metafunc.config.getoption("models"))

        metafunc.parametrize("model, scenario", scenarios)



from starling_sim.basemodel.simulation_model import SimulationModel
from starling_sim.models.SB_VS.input import Input
from starling_sim.models.SB_VS.output import Output
from starling_sim.models.SB_VS.user import User
from starling_sim.basemodel.agent.vehicles.station_based_vehicle import StationBasedVehicle
from starling_sim.basemodel.agent.stations.vehicle_sharing_station import VehicleSharingStation


class Model(SimulationModel):
    """
    Simple model for the simulation of a one-way, station-based, vehicle-sharing system
    """

    name = "Station-based vehicle-sharing model"

    agent_type_class = {
        "user": User,
        "vehicle": StationBasedVehicle,
        "station": VehicleSharingStation,
    }

    leaving_codes = {"FAIL_GET": "Failed to get a vehicle"}

    modes = {"user": ["walk"], "vehicle": ["station"], "station": [None, "walk"]}

    input_class = Input
    output_class = Output



"""
Simulation logs offer an insight of the flow of events during the simulation run.
They are displayed in chronological order and contain various information such as simulation setup,
output generation or agent activity.

Simulation logs are managed with the Python
`logging <https://docs.python.org/3.6/library/logging>`_ library.

*************************
Simulation logging levels
*************************

In addition to the base levels of *logging*, we introduce two new ones:

- An agent level that displays the agents' activities
- An algorithm level that displays the algorithms' steps

This results in the following available levels:

.. list-table:: **STARLING LOGGING LEVELS**
   :widths: auto
   :header-rows: 1
   :align: center

   * - Level
     - Numeric value
   * - CRITICAL
     - 50
   * - ERROR
     - 40
   * - WARNING
     - 30
   * - INFO
     - 20
   * - AGENT
     - 15
   * - ALGO
     - 13
   * - DEBUG
     - 10

The default logging level is 13.

To run a simulation with a different logging level, you can use the ``-l`` (or ``--level``) option of main.py.
For instance:

.. code-block:: bash

    python3 main.py data/models/SB_VS/example_nantes/ -l 20

******************
Simulation loggers
******************

We also create specific loggers for the :class:`~starling_sim.basemodel.trace.trace.Traced` and
:class:`~starling_sim.basemodel.algorithms.algorithm.Algorithm` classes with
different logging formats, in order to display additional information (such as the simulation time). For instance:

.. code-block:: text

    AGENT :: [50904], S1 : Picked up u-11

In order to display logs, instances of these classes call their
internal method log_message so the relevant information is fetched
directly from their attributes and a default level is applied.

These additional loggers may also be used by any module developed around
Starling to display simulation logs.
"""

import logging


#: logging format of the base logger
BASE_LOGGER_FORMAT = "%(levelname)s :: %(message)s"

#: default logging level of the simulations
DEFAULT_LOGGER_LEVEL = 13

#: additional logging level AGENT for agent activity logs
AGENT_LEVEL, AGENT_LEVEL_NAME = 15, "AGENT"

#: additional logging level for algorithms logs
ALGO_LEVEL, ALGO_LEVEL_NAME = 13, "ALGO"


def setup_logging(logger_level):
    """
    Setup the logging configuration for the run.

    :param logger_level: integer describing the logger level
    """

    # add levels specific to the project

    # new logging level for traced agents
    logging.addLevelName(AGENT_LEVEL, AGENT_LEVEL_NAME)

    # new logging level for algorithms
    logging.addLevelName(ALGO_LEVEL, ALGO_LEVEL_NAME)

    # set the base logger level
    logging.basicConfig(format=BASE_LOGGER_FORMAT, level=logger_level)


#: logging format of TRACED_LOGGER
TRACED_LOGGER_FORMAT = "%(levelname)s :: [%(timestamp)s], %(id)s : %(message)s"


def new_traced_logger():
    """
    Create and return a logger for the traced agents.

    :return: Logger object from the logging library
    """

    traced_logger = logging.getLogger("traced_logger")
    traced_logger.propagate = False
    formatter = logging.Formatter(TRACED_LOGGER_FORMAT)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    traced_logger.addHandler(stream_handler)

    return traced_logger


#: logging format of ALGO_LOGGER
ALGO_LOGGER_FORMAT = "%(levelname)s :: %(alg_name)s : %(message)s"


def new_algo_logger():
    """
    Create and return a logger for the simulation algorithms.

    :return: Logger object from the logging library
    """

    algo_logger = logging.getLogger("algo_logger")
    algo_logger.propagate = False
    formatter = logging.Formatter(ALGO_LOGGER_FORMAT)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    algo_logger.addHandler(stream_handler)

    return algo_logger


#: logging format of BLANK_LOGGER
BLANK_LOGGER_FORMAT = "%(message)s"


def new_blank_logger():
    """
    Create and return a logger without prefix.

    :return: Logger object from the logging library
    """

    blank_logger = logging.getLogger("blank_logger")
    blank_logger.propagate = False
    formatter = logging.Formatter(BLANK_LOGGER_FORMAT)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    blank_logger.addHandler(stream_handler)

    return blank_logger


#: Traced objects logger
TRACED_LOGGER = new_traced_logger()

#: algorithms logger
ALGO_LOGGER = new_algo_logger()

#: blank logger
BLANK_LOGGER = new_blank_logger()



from starling_sim.basemodel.output.output_factory import OutputFactory
from starling_sim.basemodel.output.kpi_output import KpiOutput
from starling_sim.basemodel.output.kpis import (
    MoveKPI,
)
from starling_sim.utils.constants import PUBLIC_TRANSPORT_TYPE


class Output(OutputFactory):
    def setup_kpi_output(self):
        # kpi output classes

        # vehicles kpi
        move_kpi = MoveKPI()
        vehicles_kpi_output = KpiOutput(PUBLIC_TRANSPORT_TYPE, [move_kpi])

        self.kpi_outputs = [
            vehicles_kpi_output,
        ]



from starling_sim.basemodel.output.output_factory import OutputFactory
from starling_sim.basemodel.output.kpi_output import KpiOutput
from starling_sim.basemodel.output.kpis import (
    MoveKPI,
    WaitKPI,
    SuccessKPI,
    DestinationReachedKPI,
    GetVehicleKPI,
    StationOccupationKPI,
    StaffOperationKPI,
    LeaveSimulationKPI,
)
from starling_sim.basemodel.output.information_factory import ActivityInformation, StockInformation


class Output(OutputFactory):
    """
    Class realising the outputs of the station-based vehicle-sharing with repositioning model
    """

    def setup_kpi_output(self):
        # kpi output classes

        # users kpi
        move_kpi = MoveKPI()
        wait_kpi = WaitKPI()
        success_kpi = SuccessKPI(["nbFailedRequest"])
        destination_kpi = DestinationReachedKPI()
        leave_simulation_kpi = LeaveSimulationKPI()
        users_kpi_output = KpiOutput(
            "user", [move_kpi, wait_kpi, success_kpi, destination_kpi, leave_simulation_kpi]
        )

        # vehicles kpi
        move_kpi = MoveKPI()
        get_vehicle_kpi = GetVehicleKPI()
        vehicles_kpi_output = KpiOutput("vehicle", [move_kpi, get_vehicle_kpi])

        # stations kpi
        success_kpi = SuccessKPI(["nbSuccessGet", "nbFailedGet", "nbSuccessPut", "nbFailedPut"])
        wait_kpi = WaitKPI()
        availability_kpi = StationOccupationKPI()
        stations_kpi_output = KpiOutput("station", [success_kpi, wait_kpi, availability_kpi])

        # staff kpi
        move_kpi = MoveKPI()
        wait_kpi = WaitKPI()
        success_kpi = SuccessKPI(["nbSuccessGet", "nbFailedGet", "nbSuccessPut", "nbFailedPut"])
        operation_kpi = StaffOperationKPI()
        staff_kpi_output = KpiOutput("staff", [move_kpi, wait_kpi, success_kpi, operation_kpi])

        self.kpi_outputs = [
            users_kpi_output,
            vehicles_kpi_output,
            stations_kpi_output,
            staff_kpi_output,
        ]

    def setup_geojson_output(self):
        super().setup_geojson_output()

        stock_factory = StockInformation()
        activity_information = ActivityInformation()

        information_factories = {
            "station": [stock_factory],
            "user": [activity_information],
            "staff": [stock_factory],
        }

        self.geojson_output.set_information_factories(information_factories)



from abc import ABC
import json


class NetworkWeight(ABC):
    """
    This class is used for computing weight on the graph edges.

    Paths on the graph are evaluated by minimising weight using
    a shortest path algorithm.
    """

    DEFAULT_PARAMETERS = {}

    def __init__(self, topology):
        self.topology = topology
        self.default_parameters = self.DEFAULT_PARAMETERS

    def pre_process_edge(self, u, v, d):
        """
        Add attributes to the edges before compute weight.

        :param u: edge origin
        :param v: edge destination
        :param d: edge data
        """
        pass

    def compute_edge_weight(self, u, v, d, parameters):
        """
        Compute and set edge weight using edge data and parameters.

        The attribute key to set is the hash of the parameters.

        :param u: edge origin
        :param v: edge destination
        :param d: edge data
        :param parameters: agent specific parameters
        """
        raise NotImplementedError()

    def get_parameters_hash(parameters):
        """
        Get the hash of the given weight parameters.

        :param parameters:

        :return: hash of the parameters, associated to the computed weight.
        """
        return str(hash(json.dumps(parameters, sort_keys=True)))

    get_parameters_hash = staticmethod(get_parameters_hash)



"""
This model describes a free-floating vehicle-sharing system.
This package contains the classes that extend the basemodel and implement the model's features.
More information on the :ref:`FF_VS` page.
"""



import simpy

from starling_sim.basemodel.agent.stations.station import Station
from starling_sim.basemodel.agent.vehicles.station_based_vehicle import StationBasedVehicle
from starling_sim.basemodel.agent.requests import StationRequest
from starling_sim.utils.utils import new_point_feature


class VehicleSharingStation(Station):
    """
    Station dedicated to a shared vehicle system
    """

    SCHEMA = {
        "properties": {
            "capacity": {
                "type": "integer",
                "title": "Station capacity",
                "description": "Maximum storing capacity of the station",
                "minimum": 0,
            },
            "stock_generation": {
                "title": "Stock generation information",
                "description": "Information for the dynamic generation of vehicles at the station. "
                "Generation is performed after the initialisation of the static inputs.",
                "type": "object",
                "properties": {
                    "generated_stock": {
                        "title": "Generated stock",
                        "description": "Generated vehicle stock of the station",
                        "type": "integer",
                        "minimum": 0,
                        "default": 0,
                    },
                    "id_format": {
                        "title": "ID format",
                        "description": "ID format of the generated agents. Placeholders: {station}, {count}",
                        "type": "string",
                        "default": "B-{station}-{count}",
                    },
                    "seats": {
                        "title": "Vehicles seats",
                        "description": "Number of seats of the generated vehicles",
                        "type": "integer",
                        "minimum": 0,
                        "default": 1,
                    },
                    "icon": {
                        "title": "Vehicles icon",
                        "description": "Icon of the generated vehicles",
                        "type": "string",
                        "default": "bike",
                    },
                },
            },
        },
        "required": ["capacity"],
    }

    def __init__(
        self, simulation_model, agent_id, origin, capacity, stock_generation=None, **kwargs
    ):
        Station.__init__(self, simulation_model, agent_id, origin, **kwargs)

        self.capacity = capacity
        self.initial_stock = 0
        self.store = simpy.Store(self.sim.scheduler.env, capacity=capacity)

        self.stock_generation = stock_generation

    def __str__(self):
        return "[id={}, position={}, capacity={}, initialStock={}]".format(
            self.id, self.position, self.capacity, self.initial_stock
        )

    def create_station_based_vehicles(self, stock_generation):
        generated_stock = stock_generation["generated_stock"]

        if generated_stock > self.capacity:
            raise ValueError(
                "Provided initial stock exceeds the station's capacity "
                "(initialising station {})".format(self.id)
            )

        # get agent_type of the model StationBasedVehicle agent
        station_based_vehicle_type = None
        for agent_type in self.sim.agent_type_class:
            if issubclass(self.sim.agent_type_class[agent_type], StationBasedVehicle):
                station_based_vehicle_type = agent_type
                break
        if station_based_vehicle_type is None:
            raise ValueError(
                "Could not find an agent class that corresponds to a station-based vehicle"
            )

        for i in range(generated_stock):
            input_dict = {
                "agent_id": stock_generation["id_format"].format(
                    station=self.id, count=self.initial_stock
                ),
                "agent_type": station_based_vehicle_type,
                "mode": self.mode,
                "icon": stock_generation["icon"],
                "seats": stock_generation["seats"],
                "station": self.id,
            }

            if self.operator is not None:
                input_dict["operator_id"] = self.operator.id

            self.sim.dynamicInput.new_agent_input(new_point_feature(properties=input_dict))

    def get_store(self):
        """
        Returns the station's store
        :return:
        """
        return self.store

    def nb_products(self):
        """
        Returns the number of products in store
        :return:
        """
        return len(self.store.items)

    def get_from_store(self, agent):
        """
        Requests a product from the station's store
        :param agent: requesting agent
        :return: StationRequest object
        """

        request = StationRequest(agent, self.sim.scheduler.now(), self, StationRequest.GET_REQUEST)

        request.set_request_event(self.store.get())

        return request

    def check(product):
        """
        Tests the type of the product according to the type of store
        This method must be extended for specific stations
        :param product: product to be tester
        :return:
        """

        return True

    check = staticmethod(check)

    def return_to_store(self, agent, product):
        """
        Tries to return a product in the station's store
        :param agent: agent returning product
        :param product: returned product
        :return: StationRequest object
        """

        if self.check(product):
            request = StationRequest(
                agent, self.sim.scheduler.now(), self, StationRequest.PUT_REQUEST
            )

            request.set_request_event(self.store.put(product))

            return request
        else:
            self.log_message(
                "Received wrong type of product ({}) from {}".format(type(product), agent.id)
            )
            return None

    def set_products(self, products):
        """
        Sets the products contained in the store
        :param products: list of products, type is not checked
        :return:
        """

        self.store.items = products

    def loop_(self):
        # generate vehicles at this station
        if self.stock_generation is not None:
            self.create_station_based_vehicles(self.stock_generation)

        yield self.execute_process(self.spend_time_())



from starling_sim.basemodel.agent.operators.station_based_operator import StationBasedOperator
from starling_sim.basemodel.algorithms.pal_zhang_dispatcher import PalZhangGreedyDispatcher


class Operator(StationBasedOperator):
    """
    This class describes the operator of the SB_VS_R model.
    """

    DISPATCHERS = {"PZ": {"punctual": PalZhangGreedyDispatcher}}

    def __init__(self, simulation_model, agent_id, fleet_dict, stations_dict, **kwargs):
        super().__init__(simulation_model, agent_id, fleet_dict, stations_dict, **kwargs)

    def loop_(self):
        # if no dispatcher is provided, do not execute repositioning
        if self.punctual_dispatcher is None:
            return

        # otherwise, follow the dispatching loop of the punctual dispatcher
        yield self.execute_process(self.punctual_dispatcher.dispatching_loop_())



"""
This module contains the MovingAgent class,
with useful methods to make an agent move in the environment
"""

from starling_sim.basemodel.agent.spatial_agent import SpatialAgent
from starling_sim.basemodel.trace.events import RouteEvent, PositionChangeEvent
from starling_sim.utils.utils import make_int
from starling_sim.utils.config import config


class MovingAgent(SpatialAgent):
    """
    This class describes an agent with the capacity of moving
    """

    def __init__(self, simulation_model, agent_id, origin, **kwargs):
        """
        Create a moving agent, setting its positional attributes and destination

        :param simulation_model:
        :param agent_id:
        :param origin:
        """

        SpatialAgent.__init__(self, simulation_model, agent_id, origin, **kwargs)

        self.tempDestination = None

    # move methods

    def move_(
        self,
        route=None,
        duration=None,
        check_dest=False,
        destination=None,
        parameters=None,
        verb=True,
    ):
        """
        Compute a route with the given parameters and follow it.

        :param route: list of consecutive nodes in the environment
        :param duration: total duration of the move
        :param check_dest: boolean indicating the destination changes should be checked while following the route
        :param destination: used for computing the route. If None, self.tempDest is used
        :param parameters: agent specific parameters used for path evaluation
        :param verb: boolean indicating if a message should be displayed when calling the function

        :return: yield one or several processes during the route execution
        """

        # get move's mode
        mode = self.mode

        # get move's destination
        if destination is None:
            destination = self.tempDestination

        # ignore moves to current position
        if destination == self.position and duration is None:
            return

        # get move's route data
        route_data = self.sim.environment.compute_route_data(
            route, duration, self.position, destination, parameters, mode
        )

        # execute route data
        yield self.execute_process(self.follow_route_data_(route_data, check_dest, mode, verb))

    def follow_route_data_(self, route_data, check_dest, mode, verb=True):
        """
        Travel along the given route

        :param route_data: route_data to be followed
        :param check_dest: boolean indicating if destination changes
         should be checked during the trip
        :param mode: mode used
        :param verb: boolean indicating if a log should be displayed when calling the function
        """

        route = route_data["route"]

        # check route

        # route departure should be self.position
        if route[0] != self.position:
            self.log_message(
                "Trying to execute a route that starts at {} while being at {}".format(
                    route[0], self.position
                ),
                40,
            )
            raise ValueError(route)
        # check route length
        if len(route) == 0:
            return

        # log agent movement
        if verb:
            self.log_message("Moving to {} with mode {}".format(route[-1], self.mode))

        # trace route move
        event = RouteEvent(self.sim.scheduler.now(), route_data, mode)
        self.trace_event(event)

        # TODO : manage if interrupted
        if check_dest:
            durations = route_data["time"]

            for i in range(1, len(route)):
                if route[-1] != self.tempDestination:
                    new_route_data = {
                        "route": route_data["route"][:i],
                        "time": route_data["time"][:i],
                        "length": route_data["length"][:i],
                    }
                    event.set_route_data(new_route_data)
                    break

                yield self.execute_process(self.move_to_(route[i], durations[i]), True)

        else:
            # execute travel and store the process
            total_time = sum(route_data["time"])
            yield self.execute_process(self.move_to_(route[-1], total_time), True)

    def change_position(self, new_position, mode):
        """
        Change the current position of the agent.

        :param new_position: new position of the agent
        :param mode: agent transport mode
        :return:
        """

        # self.trace_event(PositionChangeEvent(self.sim.scheduler.now(), new_position, mode))
        self.position = new_position

    def move_to_(self, destination, duration):
        """
        Perform the action of moving from the current position to a given destination

        :param destination: Destination in the simulation environment
        :param duration: Duration of the trip
        :return: yield a trip duration process
        """

        # wait the given duration
        yield self.execute_process(self.spend_time_(duration))

        # change agent position and trace the event
        self.change_position(destination, self.mode)

    def fly_(
        self,
        destination=None,
        duration=None,
        length=None,
        distance_factor=None,
        speed=None,
        mode=None,
        verb=True,
    ):
        """
        Realise a direct trip from current position to destination without using the network.

        The information provided (duration, length, speed) must allow to compute the
        trip duration and length. Length can be computed using euclidean distance.
        If duration and length are provided, speed is not used.

        :param destination: destination of the trip.
            Default is self.tempDestination.
        :param duration: duration of the trip.
            Default is computed using length and speed.
        :param length: length of the trip.
            Default is computed using duration and speed, or using euclidean distance.
        :param distance_factor: value multiplied to euclidean distance to get trip distance.
            Default is defined by the 'distance_factor' config value.
        :param speed: speed of the agent, used to compute duration and length
        :param mode: mode of traveling.
            Default is self.mode.
        :param verb: boolean indicating if a log should be displayed when calling the function

        :return: yield a trip duration process
        """

        # get or compute trip information
        origin = self.position

        if distance_factor is None:
            distance_factor = config["distance_factor"]

        if destination is None:
            destination = self.tempDestination

        if speed is None:
            if duration is None:
                self.log_message("Can't fly without knowing duration or speed", 40)
                raise ValueError(speed)
            else:
                if length is None:
                    length, _ = self.approximate_trip(origin, destination, distance_factor, None)

        else:
            if duration is None:
                if length is None:
                    length, duration = self.approximate_trip(
                        origin, destination, distance_factor, speed
                    )
                else:
                    duration = int(length / speed)
            else:
                if length is None:
                    length = int(duration * speed)

        if mode is None:
            mode = self.mode

        if None in [origin, destination, duration, length, mode]:
            self.log_message("Error in the fly method, missing information", 40)

        # check that duration and length are Python int
        duration = make_int(duration)
        length = make_int(length)

        # log agent movement
        if verb:
            self.log_message("Flying to {} with mode {}".format(destination, mode))

        # trace trip
        self.trace_event(
            PositionChangeEvent(
                self.sim.scheduler.now(), origin, destination, length, duration, mode
            )
        )

        # realise trip
        yield self.execute_process(self.move_to_(destination, duration))

    def approximate_trip(self, origin, destination, distance_factor, speed):
        """
        Approximate the length and time of a trip based on the euclidean distance and speed.

        The computation is based on a factor that is applied to the euclidean distance.

        :param origin: origin position
        :param destination: destination position
        :param distance_factor: value multiplied to euclidean distance to get trip distance.
        :param speed: speed of the travelling agent, in m/s.
            If None, travel time is not computed
        :return: tuple (travel_length, travel_time)
        """

        return self.sim.environment.approximate_path(
            origin, destination, distance_factor, speed, mode=self.mode
        )



from starling_sim.basemodel.algorithms.dispatcher import Dispatcher
from starling_sim.basemodel.algorithms.pal_zhang_GCH import PalZhangGCH


class PalZhangGreedyDispatcher(Dispatcher):
    SCHEMA = PalZhangGCH.SCHEMA

    def __init__(self, simulation_model, operator, verb=False):
        super().__init__(simulation_model, operator, verb)

    def init_algorithm(self):
        self.algorithm = PalZhangGCH(
            simulation_model=self.sim, operator=self.operator, verb=self.verb
        )

    def setup_dispatch(self):
        self.algorithm.setup_new_run()

    def run_algorithm(self):
        self.algorithm.run()

    def update_from_solution(self):
        self.log_message("Final planning is {}".format(self.algorithm.planning))
        self.algorithm.vehicle.set_planning(self.algorithm.planning)

    def dispatching_loop_(self):
        for time in self.algorithm.start_times:
            wait_time = time - self.sim.scheduler.now()
            yield self.sim.scheduler.timeout(wait_time)
            self.dispatch()



import logging
import sys
import hashlib
from starling_sim.basemodel.topology.osm_network import OSMNetwork
from starling_sim.basemodel.topology.empty_network import EmptyNetwork
from starling_sim.utils.config import config
from geopy import distance


class Environment:
    """
    Describes an environment in which the simulation will take place
    """

    def __init__(self, scenario, network="osm"):
        """
        Initialization of the environment, without import of data structures.

        It should be extended for the needs of the chosen environment
        """

        self.sim = None
        self.topologies = {}

        # get the topologies dict
        topologies_dict = scenario["topologies"]

        # get the 'store_paths' parameter
        if "store_paths" in scenario:
            store_paths = scenario["store_paths"]
        else:
            store_paths = False

        weight_class = None

        for mode, info in topologies_dict.items():
            # see if paths of this topology should be stored
            if isinstance(store_paths, dict):
                if mode not in store_paths:
                    raise ValueError("Missing topology in parameter 'store_paths'")
                else:
                    store = store_paths[mode]
            else:
                store = store_paths

            if info is None:
                topology = EmptyNetwork(mode, store_paths=store)
            # create a topology object according to the given network type
            elif network == "osm":
                network_file = info[0]
                speeds_file = info[1]
                if len(info) == 3:
                    weight_class = info[2]
                topology = OSMNetwork(
                    mode,
                    network_file=network_file,
                    speed_file=speeds_file,
                    store_paths=store,
                    weight_class=weight_class,
                )
            else:
                logging.error("Unknown network type {}".format(network))
                continue

            self.topologies[mode] = topology

    def setup(self, simulation_model):
        """
        Setup the environment for the simulation run.

        Set the simulation model attribute and
        prepare the environment for the simulation

        :param simulation_model: simulation model
        """

        self.sim = simulation_model
        for topology in self.topologies.values():
            topology.setup()

    # def periodic_update_(self, period):
    #     """
    #     Periodically update the simulation environment using the topology update method
    #
    #     :param period: Update period, put 0 for no updates
    #     :return:
    #     """
    #
    #     if period == 0:
    #         return
    #
    #     # Log periodic_update
    #     logging.debug("Start periodical environment update, period=" + period)
    #
    #     while period:
    #
    #         # Wait for period to elapse
    #         yield self.sim.scheduler.timeout(period)
    #
    #         # Log new update
    #         logging.debug("Environment update")
    #
    #         # Update
    #         for topology in self.topologies.values():
    #             topology.update()

    # TODO : Move to subclass ? Not very pretty
    # environment utils

    def compute_route_data(self, route, duration, origin, destination, parameters, mode):
        """
        Compute a route data from the given parameters

        :param route: list of consecutive nodes in the environment.
            If None, a shortest path from <origin> to <destination> is computed
        :param duration: total duration of the route execution.
            If None, use the "time" links of the environment
        :param origin: origin used for the shortest path computation
        :param destination: destination used for the shortest path computation
        :param parameters: agent specific parameters used for path evaluation
        :param mode: mode of the move
        :return: route_data={"route": position_list, "length": length_list, "time": time_list}
        """

        # if route is None, compute the path from <origin> to <destination> which minimises weight
        time = None
        if route is None:
            # compute shortest path
            route, time, length = self.topologies[mode].dijkstra_shortest_path_and_length(
                origin, destination, parameters
            )

        # store route in a dict
        route_data = {"route": route}

        # add length data
        self.add_route_data(route_data, mode, "length")

        # add time data
        if duration is not None:
            if duration == 0:
                speed = float("inf")
            else:
                total_length = sum(route_data["length"])
                speed = float(total_length) / duration

            self.add_route_data(route_data, mode, "time", speed=speed)
        else:
            self.add_route_data(route_data, mode, "time")

        # check that the duration fits (to avoid round-up error)
        if duration is None and time is not None:
            duration = time

        time_sum = sum(route_data["time"])

        if duration is not None and duration != time_sum:
            route_data["time"][-1] += int(duration - time_sum)

        return route_data

    def add_route_data(self, route_data, mode, data, speed=None):
        """
        Complete the given route_data with the relevant list

        :param route_data: route_data to be completed
        :param mode: mode of the move
        :param data: type of data to add, either "length" or "time"
        :param speed: speed of the move, used for "time" data.
            If None, use the links' speeds
        """

        route = route_data["route"]

        # complete a data list corresponding to the links between the route positions
        previous_position = route[0]
        data_list = [0]
        # since the link values may not be integers, we use a remainder
        remainder = 0

        for i in range(1, len(route)):
            if data == "time" and speed is not None:
                # use the speed for computing "time" data, if provided
                move_data = float(route_data["length"][i]) / speed + remainder

            else:
                # use the environment links data
                move_data = (
                    self.topologies[mode].get_edge_data(previous_position, route[i], data)
                    + remainder
                )

            # update remainder
            remainder = move_data - int(move_data)

            # add value to data list
            data_list.append(int(move_data))

            # update previous position
            previous_position = route[i]

        # update the route_data dict
        route_data[data] = data_list

    def compute_shaped_route(self, operator, shape_table, from_stop, to_stop, duration):
        # create a route_data structure
        route = [operator.stopPoints[from_stop].position]
        lengths = [0]
        length_remainder = 0
        times = [0]
        time_remainder = 0

        # complete it with the shape data
        for index, row in shape_table.iterrows():
            # append localisation to route
            if row["sequence"] == len(shape_table) + 1:
                route.append(operator.stopPoints[to_stop].position)
            else:
                route.append((row["lat"], row["lon"]))

            # append distance to lengths
            length = row["distance"] + length_remainder
            length_remainder = length - int(length)
            lengths.append(int(length))

            # append duration to times
            if row["sequence"] == len(shape_table) + 1:
                time = duration - sum(times)
            else:
                time = duration * row["distance_proportion"] + time_remainder
                time_remainder = time - int(time)
            times.append(int(time))

        # set route information
        route_data = {"route": route, "length": lengths, "time": times}

        return route_data

    def get_position(obj, position_lambda=None):
        # define lambda as obj.position if none is provided
        if position_lambda is None:

            def position_lambda(node):
                return node.position

        return position_lambda(obj)

    get_position = staticmethod(get_position)

    def get_localisation(self, node, mode=None):
        if mode is not None:
            return self.topologies[mode].position_localisation(node)
        else:
            agent_localisation = None
            for topology in self.topologies.values():
                agent_localisation = None
                try:
                    return topology.position_localisation(node)

                except KeyError:
                    pass

            return agent_localisation

    def compute_network_distance(self, source, target, mode, parameters=None, return_path=False):
        # if no mode is given, return None
        if mode is None:
            logging.warning("No mode provided for network distance computation")
            return None
        else:
            path, duration, _ = self.topologies[mode].dijkstra_shortest_path_and_length(
                source, target, parameters
            )
            if return_path:
                return path, duration
            else:
                return duration

    def compute_euclidean_distance(self, position1, position2, mode=None):
        # get position localisations
        loc1 = self.get_localisation(position1, mode)
        loc2 = self.get_localisation(position2, mode)

        # geopy would return a distance even if one localisation is None, caution
        if loc1 is None or loc2 is None:
            logging.warning(
                "One of the given localisations is None, "
                "computed euclidean distance will be false"
            )

        dist = 1000 * distance.great_circle(loc1, loc2).kilometers

        return dist

    def distance_dict_between(
        self,
        position,
        obj_list,
        distance_type,
        n=None,
        maximum_distance=None,
        position_lambda=None,
        mode=None,
        is_origin=True,
        parameters=None,
        return_path=False,
    ):
        # distance dictionary
        distance_dict = dict()

        # n can't be lesser than the list length
        if n is not None:
            n = min(n, len(obj_list))
            if n == 0:
                return distance_dict

        # compute all distances and fill the dict
        for obj in obj_list:
            if distance_type == "network":
                if is_origin:
                    distance_res = self.compute_network_distance(
                        position,
                        self.get_position(obj, position_lambda),
                        mode,
                        parameters,
                        return_path,
                    )
                else:
                    distance_res = self.compute_network_distance(
                        self.get_position(obj, position_lambda),
                        position,
                        mode,
                        parameters,
                        return_path,
                    )

            elif distance_type == "euclidean":
                distance_res = self.compute_euclidean_distance(
                    position, self.get_position(obj, position_lambda), mode
                )

            else:
                logging.warning("Unknown distance type : " + str(distance_type))
                continue

            is_too_far = False
            if maximum_distance is not None:
                if distance_type == "network" and return_path:
                    is_too_far = distance_res > maximum_distance[1]
                else:
                    is_too_far = distance_res > maximum_distance

            if not is_too_far:
                distance_dict[obj] = distance_res

        # if n is provided, only keep the n closest objects
        if n is not None:
            # sort the objects according to the length
            if distance_type == "network" and return_path:
                dict_list = sorted(list(distance_dict.keys()), key=lambda x: distance_dict[x][1])
            else:
                dict_list = sorted(list(distance_dict.keys()), key=lambda x: distance_dict[x])

            # select the objects to be removed
            dict_list = dict_list[n:]
            for obj in dict_list:
                distance_dict.pop(obj)

        # return the distance dictionary
        return distance_dict

    def euclidean_n_closest(
        self, position, obj_list, n, maximum_distance=None, position_lambda=None
    ):
        distance_dict = self.distance_dict_between(
            position,
            obj_list,
            "euclidean",
            n=n,
            maximum_distance=maximum_distance,
            position_lambda=position_lambda,
        )

        sorted_list = sorted(distance_dict.keys(), key=lambda x: distance_dict[x])

        return sorted_list

    def closest_object(
        self,
        position,
        obj_list,
        is_origin,
        mode,
        parameters=None,
        position_lambda=None,
        return_path=False,
        n=None,
    ):
        """
        Find the object of the list that is closest to the given position.

        We compute network distances between the objects of the list and the given
        position. If the n param is provided, we pre-process the list, and only
        keep the n euclidean closest objects.

        :param position:
        :param obj_list:
        :param is_origin: boolean indicating if the
        :param mode:
        :param parameters:
        :param position_lambda:
        :param return_path:
        :param n:
        :return:
        """

        if not obj_list or n == 0:
            return None

        # look only at the n (euclidean) closest objects
        if n is not None:
            obj_list = self.euclidean_n_closest(
                position=position, obj_list=obj_list, n=n, position_lambda=position_lambda
            )

        # do the network distance computation and keep the closest object
        distance_dict = self.distance_dict_between(
            position,
            obj_list,
            "network",
            position_lambda=position_lambda,
            mode=mode,
            is_origin=is_origin,
            parameters=parameters,
            return_path=return_path,
        )
        if return_path:
            closest_object = min(list(distance_dict.keys()), key=lambda x: distance_dict[x][1])
            return closest_object, distance_dict[closest_object][0]
        else:
            closest_object = min(list(distance_dict.keys()), key=lambda x: distance_dict[x])
            return closest_object

    def nearest_node_in_modes(self, localisation, modes):
        """
        Find the node common to given topologies that is closest (euclidean distance) to the localisation

        :param localisation: (lat, lng) localisation
        :param modes: transport modes, each corresponding to a topology
        :return: closest node that belongs to all topologies
        """

        # get intersection of the topologies nodes
        intersection_set = self.get_common_nodes_of(modes)

        # compute minimum euclidean distance between localisation and nodes
        best_dist = sys.maxsize
        best_node = None

        for node in intersection_set:
            # get the node's localisation
            node_localisation = self.topologies[modes[0]].position_localisation(node)

            # compute euclidean distance
            dist = distance.great_circle(localisation, node_localisation)

            # if distance is improved, keep the node
            if dist < best_dist:
                best_dist = dist
                best_node = node

        return best_node

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, modes, return_dist=False):
        """
        Call the topology localisations_nearest_nodes method.

        If a list of modes is provided, consider only nodes that are present
        in all provided topologies.

        If a node as no neighbour, assign None

        :param x_coordinates: list of X coordinates of the localisations
        :param y_coordinates: list of Y coordinates of the localisations
        :param modes: topology modes
        :param return_dist: optionally also return distance between points and nearest nodes

        :return: list of nearest nodes or (nearest_nodes, distances)
        """

        if isinstance(modes, list):
            base_graph = self.topologies[modes[0]].graph
            target_graph = base_graph.copy()

            for mode in modes[1:]:
                intersect_graph = self.topologies[mode].graph

                target_graph.remove_nodes_from(n for n in base_graph if n not in intersect_graph)

        else:
            target_graph = self.topologies[modes].graph

        # if there is no candidate nodes, the nearest node is None
        if len(target_graph.nodes) == 0:
            if isinstance(x_coordinates, list):
                nearest_nodes = [None] * len(x_coordinates)
            else:
                nearest_nodes = None
            if return_dist:
                return nearest_nodes, None
            else:
                return nearest_nodes

        target_topology = OSMNetwork(modes, graph=target_graph)

        return target_topology.localisations_nearest_nodes(
            x_coordinates, y_coordinates, return_dist=return_dist
        )

    def get_common_nodes_of(self, modes):
        """
        Find the nodes that are common to the given topologies

        :param modes: transport modes, each corresponding to a topology
        :return: set object, containing the intersection of the topology nodes
        """

        # initialize the set of common nodes with the first topology nodes
        intersection_set = set(self.topologies[modes[0]].graph.nodes)

        # realize the intersection with other topologies
        for mode in modes[1:]:
            intersection_set = intersection_set & set(self.topologies[mode].graph.nodes)

        return intersection_set

    def get_object_at(self, position, obj_list, position_lambda=None):
        """
        Return the first agent of the list that is at the given position.

        :param position:
        :param obj_list:
        :param position_lambda:
        :return:
        """

        for obj in obj_list:
            if self.get_position(obj, position_lambda) == position:
                return obj

        return None

    def approximate_path(self, origin, destination, distance_factor, speed, mode=None):
        """
        Approximate the length and time of a trip based on the euclidean distance and speed.

        The computation is based on a factor that is applied to the euclidean distance.

        :param origin: origin position
        :param destination: destination position
        :param distance_factor: value multiplied to euclidean distance to get trip distance.
        :param speed: speed of the travelling agent, in m/s.
            If None, travel time is not computed
        :param mode: position's topology.
            Default is None, then look at all environment's topologies.
        :return: tuple (travel_length, travel_time)
        """

        # compute euclidean distance between origin and destination
        euclidean_distance = self.compute_euclidean_distance(origin, destination, mode)

        # apply distance factor
        travel_length = euclidean_distance * distance_factor
        travel_length = int(travel_length)

        # apply speed (in m/s) to get travel time
        if speed is None:
            travel_time = None
        else:
            travel_time = travel_length / speed
            travel_time = int(travel_time)

        return travel_length, travel_time

    def add_node(self, node_id, properties, modes):
        """
        Add a node id and its properties to the given modes.

        :param node_id: id of the new node
        :param properties: properties of the node
        :param modes: list topologies to add the node to
        """

        for mode in modes:
            topology = self.topologies[mode]

            if node_id in topology.graph.nodes:
                # logging.warning("Adding node already in graph {}".format(node_id))
                # TODO : test if properties are same
                continue

            logging.debug("Adding node {} to topology {}".format(node_id, mode))

            topology.graph.add_node(node_id, **properties)

    def add_stops_correspondence(
        self, stops_table, modes, extend_graph, max_distance=config["max_stop_distance"]
    ):
        """
        Add a 'nearest_node' column to the given table containing the stop's nearest environment position.

        If extend_graph is True and the nearest node is further than max_distance from the stop,
        add a new node to the graph.

        :param stops_table: DataFrame with columns ["stop_id", "stop_lat", "stop_lon"]
        :param modes: list of modes
        :param extend_graph: boolean indicating if the graph should be extended when
            nearest nodes are too far
        :param max_distance: maximum node distance before extending graph
        """

        # get the stops coordinate lists
        latitudes = stops_table["stop_lat"].values
        longitudes = stops_table["stop_lon"].values

        # compute each stop nearest node
        (
            stops_table["nearest_node"],
            stops_table["node_distance"],
        ) = self.localisations_nearest_nodes(longitudes, latitudes, modes, return_dist=True)

        # extend graph with stops when needed
        # problem : with don't consider newly added nodes as possible neighbours
        if extend_graph:
            for index, row in stops_table.iterrows():
                # if the node has no close neighbour, add a new node at stop location
                if row["nearest_node"] is None or row["node_distance"] > max_distance:
                    # define node id TODO : ensure that the node doesn't already exist in topologies ?
                    m = hashlib.md5()
                    m.update(row["stop_id"].encode("utf-8"))
                    node_id = int(str(int(m.hexdigest(), 16))[0:10])

                    self.add_node(
                        node_id,
                        {"y": row["stop_lat"], "x": row["stop_lon"], "osmid": row["stop_id"]},
                        modes,
                    )

                    # update the stop's nearest node
                    stops_table.loc[index, "nearest_node"] = node_id



from starling_sim.basemodel.agent.moving_agent import MovingAgent
from starling_sim.basemodel.trace.events import (
    MoveEvent,
    RouteEvent,
    PositionChangeEvent,
    GetVehicleEvent,
    LeaveVehicleEvent,
)


class Vehicle(MovingAgent):
    """
    This class describes the basic features of a vehicle agent

    It should be extended to implement more specific features and behaviours
    """

    SCHEMA = {
        "properties": {
            "seats": {
                "type": "integer",
                "title": "Number of seats of the vehicle",
                "description": "Capacity of the vehicle",
                "minimum": 0,
            }
        },
        "required": ["seats"],
    }

    def __init__(self, simulation_model, agent_id, origin, seats, **kwargs):
        """
        A vehicle have a number of seats and a list of occupants.

        :param simulation_model:
        :param agent_id:
        :param origin:
        :param seats:
        :param kwargs:
        """

        MovingAgent.__init__(self, simulation_model, agent_id, origin, **kwargs)

        self.seats = seats
        self.occupants = []
        self.vehicle = None

    def __str__(self):
        return "[id={}, origin={}, seats={}]".format(self.id, self.origin, self.seats)

    def change_position(self, new_position, mode):
        """
        The vehicle changes the position of all of its occupants along with its own.

        :param new_position:
        :param mode:
        :return:
        """

        # change occupants positions
        for occupant in self.occupants:
            occupant.change_position(new_position, mode)

        # change own position
        super().change_position(new_position, mode)

    def trace_event(self, event):
        """
        The vehicle adds its move and route traces to its occupants.

        :param event:
        :return:
        """

        # add event to occupants traces
        if (
            isinstance(event, MoveEvent)
            or isinstance(event, RouteEvent)
            or isinstance(event, PositionChangeEvent)
        ):
            for occupant in self.occupants:
                occupant.trace_event(event)

        # add event to own trace
        super().trace_event(event)

    def load(self):
        """
        Compute the current load of the vehicle.

        :return: vehicle load
        """

        load = 0
        for occupant in self.occupants:
            if hasattr(occupant, "number"):
                load += occupant.number
            else:
                load += 1

        return load

    def add_passenger(self, passenger):
        if self.load() >= self.seats:
            self.log_message("Adding a new passenger while full", 30)

        passenger.vehicle = self
        self.occupants.append(passenger)

    def get_passenger(self, passenger):
        """
        Add the passenger to the list of occupants.

        :param passenger: agent
        """

        # add passenger
        self.add_passenger(passenger)

        # trace event
        passenger.trace_event(GetVehicleEvent(self.sim.scheduler.now(), passenger, self))

    def remove_passenger(self, passenger):
        # check if passenger is in vehicle
        if passenger not in self.occupants is None:
            self.log_message("Tried to leave passenger not in vehicle", 30)
            return

        # leave passenger
        self.occupants.remove(passenger)
        passenger.vehicle = None

    def leave_passenger(self, passenger):
        """
        Remove the passenger from the list of occupants.

        :param passenger: agent
        """

        # remove passenger from occupants
        self.remove_passenger(passenger)

        # trace event
        passenger.trace_event(LeaveVehicleEvent(self.sim.scheduler.now(), passenger, self))



from starling_sim.basemodel.agent.moving_agent import MovingAgent
from starling_sim.basemodel.trace.events import (
    RequestEvent,
    GetVehicleEvent,
    LeaveVehicleEvent,
    DestinationReachedEvent,
)
from starling_sim.utils.constants import SUCCESS_LEAVE

import sys


class Person(MovingAgent):
    """
    This class describes the basic features of a human agent

    It should be extended to implement more specific features and behaviours
    """

    SCHEMA = {
        "properties": {
            "mode": {
                "type": "string",
                "title": "Default network",
                "description": "Road network used by the agent",
                "default": "walk",
            },
            "destination": {
                "type": ["number", "string"],
                "title": "Destination position",
                "description": "Destination position id (inferred from geometry)",
            },
            "origin_time": {
                "type": "integer",
                "title": "Activity start time [seconds]",
                "description": "Time at which the agent will enter the simulation",
                "minimum": 1,
            },
            "number": {
                "title": "User group load",
                "description": "Integer indicating if the agent corresponds to a "
                "group of several persons, and will thus occupy several seats",
                "type": "integer",
                "default": 1,
                "minimum": 0,
            },
            "max_tries": {
                "advanced": True,
                "title": "Maximum number of service tryouts",
                "description": "Number of failed attempts after which the agent will leave the simulation."
                " If not specified, the number of tries is infinite.",
                "type": ["integer", "null"],
                "minimum": 0,
                "default": None,
            },
            "fail_timeout": {
                "advanced": True,
                "title": "Request fail timeout [seconds]",
                "description": "Time waited when a request fails",
                "type": "integer",
                "minimum": 0,
                "default": 0,
            },
        },
        "required": ["destination", "origin_time"],
    }

    def __init__(
        self, simulation_model, agent_id, origin, destination, origin_time, number=1, **kwargs
    ):
        """
        Initialize the new Person object, as a moving agent with specific user data.

        :param simulation_model:
        :param agent_id:
        :param origin: Origin position of the agent
        :param destination: Destination position of the agent
        :param origin_time: time at which the person should enter the simulation
        :param number: user group load
        :param max_tries: maximum number of failed system tries before leaving the system.
            Default is None (infinite tries).
        :param fail_timeout: time waited when a request fails
        :param kwargs:
        """

        super().__init__(simulation_model, agent_id, origin, **kwargs)

        # destination position of the agent, its spatial objective
        self.destination = destination

        # time at which the person should enter the simulation
        self.originTime = origin_time

        # user group load (indicates if the agent corresponds to several persons)
        self.number = number

        # number of failed system tries before leaving. If None, try indefinitely.
        self.maxTries = self.profile["max_tries"]

        # time to wait when a request fails
        self.failTimeout = self.profile["fail_timeout"]

        # persons start with no vehicle
        self.vehicle = None

    def __str__(self):
        return "[id={}, origin={}, destination={}, vehicle={}]".format(
            self.id, self.origin, self.destination, self.vehicle
        )

    def trace_event(self, event):
        """
        Add the event to the person's trace and to other relevant agents.

        In the case of a movement using a vehicle, the vehicle itself traces
        the movements of the agent.

        :param event: Event object to be added to the person's trace
        """

        # add event to agent trace
        super().trace_event(event)

        # add event to agent's vehicle trace
        try:
            if self.vehicle is not None:
                # relevant events for vehicle (move and position already called)
                if isinstance(event, GetVehicleEvent) or isinstance(event, LeaveVehicleEvent):
                    self.vehicle.trace_event(event)

        except AttributeError:
            # in case of generation event, agent don't have vehicle attribute already
            pass

        # add request events to requested agents
        if isinstance(event, RequestEvent):
            if event.request.structure is not None:
                event.request.structure.trace_event(event)

    # simulation interactions

    def move_shortest_with_vehicle_(self, destination=None, duration=None, parameters=None):
        """
        Call self.vehicle's move method.

        :param destination: agent destination, default is self.tempDestination
        :param duration: total duration of the move
        :param parameters: agent specific parameters used for path evaluation
        :return: yield the moving process of the agent's vehicle
        """

        # use the person's tempDest
        if destination is None:
            destination = self.tempDestination

        if self.vehicle is None:
            self.log_message("Tried to travel with vehicle but has None", 30)
            return
        else:
            yield self.execute_process(
                self.vehicle.move_(
                    destination=destination, duration=duration, parameters=parameters
                )
            )

    def walk_to_destination_(self):
        """
        Walk from current position to the destination position.

        :return:
        """
        # update tempDest
        self.tempDestination = self.destination

        # move to destination
        yield self.execute_process(self.move_())

        # trace destination reached event
        self.trace_event(DestinationReachedEvent(self.sim.scheduler.now()))

        # leave simulation with successful leave code
        self.leave_simulation(SUCCESS_LEAVE)

    def get_vehicle(self, vehicle):
        """
        Get the given vehicle.

        Set the agent's vehicle attribute, and add the agent to the
        vehicle occupants. This method should only be called after
        making sure the vehicle is available

        :param vehicle: Vehicle object
        """

        # get vehicle
        self.vehicle = vehicle
        vehicle.occupants.append(self)

        # trace event
        self.trace_event(GetVehicleEvent(self.sim.scheduler.now(), self, vehicle))

    def leave_vehicle(self):
        """
        Leave the currently owned vehicle.

        Set the agent's vehicle attribute, and add the agent to the
        vehicle occupants. This method should only be called after
        making sure the owner is allowed to leave the vehicle.
        """

        # check if agent has a vehicle
        if self.vehicle is None:
            self.log_message("Tried to leave vehicle without owning one", 30)
            return None

        # trace event
        self.trace_event(LeaveVehicleEvent(self.sim.scheduler.now(), self, self.vehicle))

        # leave vehicle
        self.vehicle.occupants.remove(self)
        self.vehicle = None

    def closest_walkable_node_of(
        self, mode, position=None, n=1, parameters=None, return_path=False
    ):
        """
        Compute the closest node of the given mode that can be reached by walking.

        The node is computed by looking at the common nodes of the two modes. The default
        behaviour is then to look at the euclidean closest one (with n=1), but the shortest
        path can also be used. It can be returned with the result node.

        :param mode: mode that should be reached by walking
        :param position: departure position.
            Default is self.position.
        :param n: simplify the computation to the n euclidean closest nodes.
            Default is 1, we look at the euclidean closest node .
        :param parameters: agent parameters for computing weight
        :param return_path: boolean indicating if the path should be returned.
            Default is False.
        :return: closest_node, or (closest_node, path) if return_path
        """

        # default position is self.position
        if position is None:
            position = self.position

        # get common nodes of the given modes
        intersection_set = self.sim.environment.get_common_nodes_of(["walk", mode])

        result = self.sim.environment.closest_object(
            position,
            intersection_set,
            True,
            "walk",
            parameters=parameters,
            position_lambda=lambda x: x,
            return_path=return_path,
            n=n,
        )

        return result

    # request methods
    # TODO : put these in a client class ?

    def try_new_request_(self):
        """
        Realise a new request according to the agent state.

        This method must be implemented according to the user's
        request mode

        :return: Request object, completed according to request result
        """

        pass

    def request_loop_(self, max_tentatives=None):
        """
        Loop over the requests and return the last one.

        This method allows the user to loop over a series of requests. The user
        either stop after succeeding in a request, or after reaching the
        maximum number of tentatives (which defaults to infinite). The creation
        and execution of new requests are realised by calling the try_new_request
        method, which must be implemented for the needs of the model.

        :param max_tentatives: number of failed requests until giving up
        :return: last Request object, completed according to request result
        """

        new_request = None
        success = False
        request_tentatives = 0

        if max_tentatives is None:
            max_tentatives = sys.maxsize

        while not success and request_tentatives < max_tentatives:
            # create a new request and try it
            new_request = yield self.execute_process(self.try_new_request_())

            # trace request event
            request_event = RequestEvent(self.sim.scheduler.now(), new_request)
            self.trace_event(request_event)

            # update request tentative and success
            request_tentatives += 1
            success = new_request.success

            # wait a bit
            if self.failTimeout is not None and not new_request.success:
                yield self.execute_process(self.spend_time_(self.failTimeout))

        return new_request

    def wait_for_request_(self, request, patience=None):
        """
        Wait for the given request, respecting a patience threshold.

        :param request: Request object
        :param patience: patience threshold, None corresponds to infinite patience
        :return: yield the request event, with a timeout event if patience is provided,
         and returns the request result
        """

        if patience is not None:
            result = yield request.event_ | self.sim.scheduler.timeout(patience)
        else:
            result = yield request.event_
            result = {request.event_: result}

        return result

    def wait_trip_request_(self, request):
        # wait for the pickup
        yield request.pickupEvent_

        if self.vehicle is None:
            request.cancel()
            return False

        # wait for the dropoff
        yield request.dropoffEvent_

        if self.vehicle is not None:
            self.simulation_error("Has vehicle after dropoff")

        return True

    # TODO : request process unification

    def station_get_request(self, local_station):
        """
        Request (GET) the station at current position.

        :param local_station: Station object at current position
        :return: StationRequest object
        """

        # check position
        if self.position != local_station.position:
            self.log_message("Tried to request distant station", 40)

        # request vehicle
        request = local_station.get_from_store(self)

        self.log_message("Made a request : {}".format(request))

        return request

    def station_put_request(self, local_station):
        """
        Request (PUT) the station at current position.

        :param local_station: Station object at current position
        :return: StationRequest object
        """

        # check position
        if self.position != local_station.position:
            self.log_message("Tried to request distant station", 40)

        # request vehicle
        request = local_station.return_to_store(self, self.vehicle)

        self.log_message("Made a request : {}".format(request))

        return request



"""
This module manages the parameters of the simulation
"""

import datetime
import logging
import os
import json
from copy import deepcopy

from starling_sim.utils.utils import json_load, add_defaults_and_validate
from starling_sim.utils import paths
from starling_sim.utils.config import config
from starling_sim import __version__


class SimulationScenario:
    """
    This class describes a simulation scenario folders and data.
    """

    BASE_PARAM_SCHEMA = "parameters.schema.json"

    def __init__(self, scenario_folder_path: str):
        # scenario model code
        self.model = None

        # scenario name
        self.name = None

        # simulation parameters (can be accessed using SimulationScenario["key"])
        self.parameters = None

        # run summary
        self.runSummary = None

        # scenario folder path
        self.scenario_folder = None

        # scenario inputs folder path
        self.inputs_folder = None

        # scenario outputs folder path
        self.outputs_folder = None

        # set the scenario folders
        self._set_scenario_folders(scenario_folder_path)

    def _set_scenario_folders(self, scenario_folder_path: str):
        """
        Set the folder attributes from the scenario folder path.

        Paths are built according to the structure enforced by the functions of starling_sim.utils.paths.

        :param scenario_folder_path: path to the scenario folder
        """

        # check scenario folder
        self.scenario_folder = os.path.join(scenario_folder_path, "")
        if not (os.path.exists(self.scenario_folder) and os.path.isdir(self.scenario_folder)):
            raise ValueError(
                "The scenario folder does not exist or is not a folder : {}".format(
                    self.scenario_folder
                )
            )

        # check inputs folder
        self.inputs_folder = paths.scenario_inputs_folder(self.scenario_folder)
        if not (os.path.exists(self.inputs_folder) and os.path.isdir(self.inputs_folder)):
            raise ValueError(
                "The inputs folder does not exist or is not a folder : {}".format(
                    self.inputs_folder
                )
            )

        # create outputs folder if it does not exist
        if "OUTPUT_FOLDER" in os.environ:
            output_folder = os.path.join(os.environ["OUTPUT_FOLDER"], "")
        else:
            output_folder = paths.scenario_outputs_folder(self.scenario_folder)
        self.outputs_folder = output_folder
        if not os.path.exists(self.outputs_folder):
            os.mkdir(self.outputs_folder)

    def get_scenario_parameters(self):
        """
        Get and validate the scenario simulation parameters.

        Also get the scenario model and name from the parameters.

        The parameter file path is enforced by the scenario_parameters_filepath function.
        """

        try:
            parameters_path = paths.scenario_parameters_filepath(self.scenario_folder)
            self.parameters = json_load(parameters_path)

        except (FileNotFoundError, json.JSONDecodeError) as e:
            raise ValueError("Error while loading the scenario parameters : {}".format(str(e)))

        # check scenario folder against scenario name
        # in fact this is not necessary, scenario name is not used for pathing anymore
        if os.path.basename(os.path.dirname(self.scenario_folder)) != self.parameters["scenario"]:
            raise ValueError("The scenario folder must be named after the scenario name")

        # parameters validation
        schema = json_load(paths.schemas_folder() + self.BASE_PARAM_SCHEMA)
        self.parameters = add_defaults_and_validate(self.parameters, schema)

        # change date format from YYYY-MM-DD to YYYYMMDD
        if "date" in self.parameters:
            self.parameters["date"] = self.parameters["date"].replace("-", "")

        # get model and scenario
        self.model = self.parameters["code"]
        self.name = self.parameters["scenario"]

        self.init_run_summary()

    def __getitem__(self, item):
        """
        Method called when using 'SimulationScenario[item]'

        :param item: Name of the parameter accessed
        :return: self.parameters[item]
        """

        if item not in self.parameters:
            logging.error(
                "Trying to access unknown parameter '"
                + str(item)
                + "'. Does it appears in the parameter entry ?"
            )
            raise KeyError("The parameter '" + str(item) + "' does not exist")

        return self.parameters[item]

    def __contains__(self, item):
        """
        Method called when using 'item in simulationParameters'

        :param item:
        :return: True if item is in self.parameters, False otherwise
        """

        return item in self.parameters

    def init_run_summary(self):
        """
        Initialise the run summary.
        """

        summary = dict()

        # get run date
        summary["date"] = str(datetime.datetime.today())

        # get starling version
        summary["starling_version"] = __version__

        # copy scenario parameters
        summary["parameters"] = self.copy_parameters()

        # copy config
        summary["config"] = config.copy()

        # scenario output files
        summary["output_files"] = dict()

        # run statistics
        summary["stats"] = dict()

        self.runSummary = summary

    def copy_parameters(self):
        """
        Return a deepcopy of the simulation parameters.

        :return: deepcopy of simulation parameters
        """

        return deepcopy(self.parameters)

    def set_stat(self, key, value):
        """
        Set a value for the given stat key.

        :param key: stat key
        :param value: stat value
        """
        stats = self.runSummary["stats"]

        if key in stats:
            logging.warning("Overwriting '{}' scenario statistic".format(key))

        stats[key] = value



from starling_sim.basemodel.output.geojson_output import new_geojson_output
from starling_sim.utils.utils import json_pretty_dump, create_file_information
from starling_sim.utils.config import config
from starling_sim.utils.constants import RUN_SUMMARY_FILENAME

import logging


class OutputFactory:
    """
    Describes an output generation method

    This class should be extended to give a concrete implementation of the output generator
    e.g. writing a json containing all the simulation data
    """

    GENERATION_ERROR_FORMAT = "Error while generating {} output: {}"

    def __init__(self):
        """
        The constructor must be extended for the needs of the generation method
        """

        # list of output files and associated information
        self.output_files = []

        # list of KpiOutput objects, each will generate one kpi file
        self.kpi_outputs = None

        # GeojsonOutput object, will generate the visualisation output
        self.geojson_output = None

        self.sim = None

    def setup(self, simulation_model):
        """
        Setup method called during simulation setup.

        Set values of the output factory attributes.

        :param simulation_model:
        """

        # set simulation model
        self.sim = simulation_model

        # get output folder
        output_folder = simulation_model.scenario.outputs_folder

        # get scenario
        scenario = simulation_model.scenario.name

        # setup kpi outputs

        self.setup_kpi_output()

        for kpi_output in self.kpi_outputs:
            # build the kpi output filename
            kpi_filename = config["kpi_format"].format(
                scenario=scenario, kpi_output=kpi_output.name
            )

            # set kpi output file
            kpi_output.setup(kpi_filename, output_folder, simulation_model)

        # setup geojson output

        self.setup_geojson_output()

        if self.geojson_output is not None:
            # build the geojson output filename
            geojson_filename = config["geojson_format"].format(scenario=scenario)

            self.geojson_output.setup(simulation_model, geojson_filename, output_folder)

    def setup_kpi_output(self):
        """
        Set the kpi_outputs attribute as a list of KpiOutput objects.

        By default, no KPIs are set
        """

        self.kpi_outputs = []

    def setup_geojson_output(self):
        """
        Set the geojson_output attribute as a GeojsonOutput object.
        """

        self.geojson_output = new_geojson_output()

    def new_output_file(
        self,
        filepath: str,
        mimetype: str,
        compressed_mimetype: str = None,
        content: str = None,
        subject: str = None,
    ):
        """
        Add a new file and its information to the output dict.

        This method should be called after generating an output file.

        :param filepath: output file path
        :param mimetype: file mimetype
        :param compressed_mimetype: compressed mimetype (defaults to file mimetype)
        :param content: content metadata (mandatory)
        :param subject: subject
        """

        output_file_information = create_file_information(
            filepath,
            mimetype,
            compressed_mimetype=compressed_mimetype,
            content=content,
            subject=subject,
        )

        logging.info(
            "Generated {} output in file {}".format(
                output_file_information["metadata"]["content"], filepath
            )
        )

        self.output_files.append(output_file_information)

    def extract_simulation(self, simulation_model):
        """
        This method will be called for the output generation.

        It must be extended to generate the output using specific methods.
        """

        # traces output
        if simulation_model.scenario["traces_output"]:
            try:
                self.generate_trace_output(simulation_model)
            except Exception as e:
                logging.warning(self.GENERATION_ERROR_FORMAT.format("traces", e))

        # kpi output
        if simulation_model.scenario["kpi_output"]:
            self.generate_kpi_output(simulation_model)

        # geojson output

        if simulation_model.scenario["visualisation_output"]:
            try:
                self.generate_geojson_output(simulation_model)
            except Exception as e:
                logging.warning(self.GENERATION_ERROR_FORMAT.format("visualisation", e))

        # run summary output
        self.generate_run_summary(simulation_model.scenario)

    def generate_geojson_output(self, simulation_model):
        """
        Call the generation method of the geojson output attribute

        :param simulation_model:
        """

        self.geojson_output.add_population_features()

        self.geojson_output.generate_geojson()

    def generate_kpi_output(self, simulation_model):
        """
        Call the generation method of all the kpi outputs

        :param simulation_model:
        """

        for kpi_output in self.kpi_outputs:
            try:
                kpi_output.write_kpi_table()
            except Exception as e:
                logging.warning(self.GENERATION_ERROR_FORMAT.format(kpi_output.name + " kpi", e))

    def generate_trace_output(self, simulation_model):
        """
        Generate a text file containing the event traces of the agents.

        :param simulation_model:
        """

        # get the scenario information and outfile
        scenario = simulation_model.scenario.name
        model_code = simulation_model.scenario.model
        output_folder = simulation_model.scenario.outputs_folder
        filepath = output_folder + config["traces_format"].format(scenario=scenario)

        # open the trace file in write mode
        with open(filepath, "w") as outfile:
            # write header with scenario information
            outfile.write("Traces of scenario {}, model {}".format(scenario, model_code))

            # write the trace of the dynamic input
            outfile.write("\nTrace of dynamicInput")
            for event in simulation_model.dynamicInput.trace.eventList:
                outfile.write("\n")
                outfile.write(str(event))

            # write the trace of the simulation agents
            for agent in simulation_model.agentPopulation.get_total_population():
                # don't display agents with empty trace
                if len(agent.trace.eventList) <= 2:
                    continue

                outfile.write("\nTrace of agent " + str(agent.id))
                for event in agent.trace.eventList:
                    outfile.write("\n")
                    outfile.write(str(event))

        self.sim.outputFactory.new_output_file(filepath, "text/plain", content="traces")

    def generate_run_summary(self, simulation_scenario):
        """
        Generate a summary file of the simulation run.

        :param simulation_scenario: SimulationScenario object
        """
        filepath = simulation_scenario.outputs_folder + RUN_SUMMARY_FILENAME

        # add run summary to output files
        self.sim.outputFactory.new_output_file(filepath, "application/json", content="run_summary")

        # set output files in run summary and dump it in a file
        simulation_scenario.runSummary["output_files"] = self.output_files
        json_pretty_dump(simulation_scenario.runSummary, filepath)



from starling_sim.basemodel.trace.events import *
from starling_sim.basemodel.agent.requests import UserStop, StopPoint, StationRequest
from starling_sim.utils.constants import (
    SERVICE_INIT,
    SERVICE_UP,
    SERVICE_PAUSE,
    SERVICE_END,
)
import inspect
from abc import ABC, abstractmethod
import math


class KPI(ABC):
    """
    Generic structure of a KPI class

    Its sub-classes compute and update specific indicator
    from given events
    """

    # indicates if this KPI is compatible with time profiling
    PROFILE_COMPATIBILITY = True

    #: **agentId**: id of the agent
    KEY_ID = "agentId"

    def __init__(self, export_keys: list = None):
        """
        The indicator_dict associates values to the indicators names

        The keys attribute correspond to the keys of the indicator_dict
        """

        self.kpi_output = None

        # list of indicator names of this KPI
        self.keys = []
        self._export_keys = export_keys

        # current agent on which KPI is evaluated
        self.agent = None

        # event time follow up
        self.current_timestamp = None

        # indicators evaluation attributes
        self.indicator_dict = None

        # time profiling
        self.profile = None
        self.current_profile_index = 0

    @property
    def export_keys(self):
        return self._export_keys if self._export_keys is not None else self.keys

    def _init_keys(self):
        # return class attributes starting with "KEY_"
        keys = list(
            map(
                lambda x: x[1],
                filter(
                    lambda x: x[0].startswith("KEY_"),
                    inspect.getmembers(self.__class__, lambda a: not (inspect.isroutine(a))),
                ),
            )
        )
        keys.remove(self.KEY_ID)
        return keys

    def new_indicator_dict(self):
        """
        Evaluate indicators starting values

        :return: dict with keys in self.keys
        """
        return {key: 0 for key in self.keys}

    def setup(self, kpi_output, simulation_model):
        """
        Setup method called during simulation setup.

        After calling this method, the `keys` attribute should be set.

        :param kpi_output: parent KpiOutput
        :param simulation_model:
        """
        self.kpi_output = kpi_output
        time_profile = self.kpi_output.sim.scenario["kpi_time_profile"]
        if self.PROFILE_COMPATIBILITY and time_profile:
            if isinstance(time_profile, bool):
                # default profile is one hour intervals until simulation ends
                self.profile = [
                    hour * 3600
                    for hour in range(math.ceil(self.kpi_output.sim.scenario["limit"] / 3600))
                ]
            else:
                # otherwise, use provided time profile (add missing 0)
                self.profile = time_profile if time_profile[0] == 0 else [0] + time_profile

        self._indicators_setup(simulation_model)
        self.keys = self._init_keys()

    def _indicators_setup(self, simulation_model):
        """
        Setup attributes that need the simulation model to be initialised.
        """
        pass

    def evaluate_for_agent(self, agent):
        """
        Evaluate KPI indicators for the given agent.

        Indicators are evaluated by browsing the agent's
        events in chronological and updating values according
        to the KPI description.

        :param agent: Traced agent
        """
        # reset kpi
        self.reset_for_agent(agent)

        # browse agent's events in chronological order
        events = sorted(agent.trace.eventList, key=lambda x: x.timestamp)
        for event in events:
            self.update_from_event(event)

        # signal end of events
        self.end_of_events()

    def reset_for_agent(self, agent):
        """
        Reset indicators of the KPI in preparation of a new evaluation.

        :param agent:
        """
        # set studied agent
        self.agent = agent
        # reset indicators attributes
        self.current_timestamp = 0
        self.indicator_dict = self.new_indicator_dict()
        self.current_profile_index = 0

    def end_of_events(self):
        """
        Execute certain actions when the event list evaluation has ended.
        """
        if self.profile:
            # if profiling is on, fill rows until the last profile interval
            while self.current_profile_index < len(self.profile):
                self.end_of_profile_range()
        else:
            # otherwise, just add a row containing the KPI evaluation
            self.new_kpi_row()

    def end_of_profile_range(self):
        """
        Add a new row with indicators of current interval and jump to the next one.
        """
        self.current_profile_index += 1
        self.new_kpi_row()

    def new_kpi_row(self):
        """
        Add a new row to the output table.

        Indicators are reset after their data has been added to a row.
        """
        # add kpi row
        for key in self.export_keys:
            self.kpi_output.kpi_rows[key].append(self.indicator_dict[key])

        # reset indicators
        self.indicator_dict = self.new_indicator_dict()

    def is_in_later_profile(self, timestamp):
        """
        Test if time profiling is enabled and if the given timestamp is in a later profile interval.

        :param timestamp:
        :return: boolean indicating if timestamp in a later profile interval
        """
        return (
            self.profile
            and self.current_profile_index < len(self.profile) - 1
            and timestamp >= self.profile[self.current_profile_index + 1]
        )

    def update_from_event(self, event: Event):
        """
        Update indicators based on the event contents.

        If event is in a later profile interval, jump
        to the right interval before processing the event.

        :param event: Event instance
        """
        assert (
            event.timestamp >= self.current_timestamp
        ), "Event list should be ordered chronologically"

        # if event is in a later profile range, jump to it
        while self.is_in_later_profile(event.timestamp):
            self.end_of_profile_range()

        # update timestamp
        self.current_timestamp = event.timestamp

        self._update(event)

    def add_proportioned_indicators(self, event):
        # evaluate total duration of event
        if not isinstance(event, DurationEvent):
            raise ValueError(
                "add_proportioned_indicators should be called on DurationEvent instances"
            )
        total_duration = event.total_duration

        # while current profile does not contain profile end, add proportioned indicators and jump to next profile
        current_timestamp = event.timestamp
        while self.is_in_later_profile(event.timestamp + total_duration):
            # evaluate the duration spent in the current profile
            duration_current = self.profile[self.current_profile_index + 1] - current_timestamp

            # evaluate indicators for the current profile range
            self.evaluate_indicators_on_profile_range(event, current_timestamp, duration_current)

            self.end_of_profile_range()
            current_timestamp = self.profile[self.current_profile_index]

        duration_last = event.timestamp + total_duration - current_timestamp
        self.evaluate_indicators_on_profile_range(event, current_timestamp, duration_last)

    def evaluate_indicators_on_profile_range(self, event, current_timestamp, duration_on_range):
        """
        Evaluate and update KPI indicators for the given event on the specified time interval.

        :param event:
        :param current_timestamp:
        :param duration_on_range:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def _update(self, event):
        """
        Update the kpi values according to the event content and the agent.

        :param event: processed event
        :return:
        """


class MoveKPI(KPI):
    """
    This KPI evaluates the distance and spent time for each one of the simulation modes
    """

    #: **{mode}Distance**: distance travelled in <mode> [meters]
    SUFFIX_KEY_DISTANCE = "{mode}Distance"
    #: **{mode}Time**: time travelled in <mode> [seconds]
    SUFFIX_KEY_TIME = "{mode}Time"

    def __init__(self, **kwargs):
        self.modes = []
        super().__init__(**kwargs)

    def _indicators_setup(self, simulation_model):
        self.modes = list(simulation_model.environment.topologies.keys())

    def _init_keys(self):
        keys = []
        for mode in self.modes:
            keys.append(self.SUFFIX_KEY_DISTANCE.format(mode=mode))
            keys.append(self.SUFFIX_KEY_TIME.format(mode=mode))
        return keys

    def _update(self, event):
        if isinstance(event, MoveEvent):
            self.add_proportioned_indicators(event)

    def evaluate_indicators_on_profile_range(self, event, current_timestamp, duration_on_range):
        if duration_on_range == 0:
            duration = 0
            distance = event.distance
        else:
            if isinstance(event, RouteEvent):
                route_data = event.get_route_data_in_interval(
                    current_timestamp, current_timestamp + duration_on_range
                )
                duration = sum(route_data["time"])
                distance = sum(route_data["length"])
            else:
                duration = duration_on_range
                distance = round(event.distance * duration_on_range / event.duration)
        self.indicator_dict[self.SUFFIX_KEY_TIME.format(mode=event.mode)] += duration
        self.indicator_dict[self.SUFFIX_KEY_DISTANCE.format(mode=event.mode)] += distance


class WaitKPI(KPI):
    """
    This KPI evaluates the time spent waiting
    """

    #: **waitTime**: total traced wait time [seconds]
    KEY_WAIT = "waitTime"

    def _update(self, event):
        if isinstance(event, WaitEvent) or isinstance(event, RequestEvent):
            self.add_proportioned_indicators(event)

    def evaluate_indicators_on_profile_range(self, event, current_timestamp, duration_on_range):
        self.indicator_dict[self.KEY_WAIT] += duration_on_range


class GetVehicleKPI(KPI):
    """
    This KPI evaluates the number of vehicle uses
    """

    #: **nbGetVehicle**: number of uses of the vehicle
    KEY_GET_VEHICLE = "nbGetVehicle"

    def _update(self, event):
        """
        Add a new use for each GetVehicleEvent

        :param event:
        """

        if isinstance(event, GetVehicleEvent):
            self.indicator_dict[self.KEY_GET_VEHICLE] += event.agent.number


class SuccessKPI(KPI):
    """
    This KPI evaluates the number of failed/successful requests
    """

    #: **nbFailedGet**: number of failed get requests
    KEY_FAILED_GET = "nbFailedGet"
    #: **nbSuccessGet**: number successful get requests
    KEY_SUCCESS_GET = "nbSuccessGet"
    #: **nbFailedPut**: number of failed put requests
    KEY_FAILED_PUT = "nbFailedPut"
    #: **nbSuccessPut**: number of successful put requests
    KEY_SUCCESS_PUT = "nbSuccessPut"
    #: **nbFailedRequest**: number of failed requests
    KEY_FAILED_REQUEST = "nbFailedRequest"
    #: **nbSuccessRequest**: number of successful requests
    KEY_SUCCESS_REQUEST = "nbSuccessRequest"

    def _init_keys(self):
        return [
            self.KEY_FAILED_GET,
            self.KEY_SUCCESS_GET,
            self.KEY_FAILED_PUT,
            self.KEY_SUCCESS_PUT,
            self.KEY_FAILED_REQUEST,
            self.KEY_SUCCESS_REQUEST,
        ]

    def _update(self, event):
        """
        Add request events according to their success

        :param event:
        """

        if isinstance(event, RequestEvent):
            if event.request.success:
                self.indicator_dict[self.KEY_SUCCESS_REQUEST] += 1
                if event.request.type == StationRequest.GET_REQUEST:
                    self.indicator_dict[self.KEY_SUCCESS_GET] += 1
                else:
                    self.indicator_dict[self.KEY_SUCCESS_PUT] += 1
            else:
                self.indicator_dict[self.KEY_FAILED_REQUEST] += 1
                if event.request.type == StationRequest.GET_REQUEST:
                    self.indicator_dict[self.KEY_FAILED_GET] += 1
                else:
                    self.indicator_dict[self.KEY_FAILED_PUT] += 1


class StaffOperationKPI(KPI):
    """
    This KPI evaluates the number of staff operations
    """

    #: **nbFailedGetStaff**: number of failed gets by staff
    KEY_FAILED_GET_STAFF = "nbFailedGetStaff"
    #: **nbSuccessGetStaff**: number of successful gets by staff
    KEY_SUCCESS_GET_STAFF = "nbSuccessGetStaff"
    #: **nbFailedPutStaff**: number of failed puts by staff
    KEY_FAILED_PUT_STAFF = "nbFailedPutStaff"
    #: **nbSuccessPutStaff**: number of successful puts by staff
    KEY_SUCCESS_PUT_STAFF = "nbSuccessPutStaff"

    def _init_keys(self):
        return [
            self.KEY_FAILED_GET_STAFF,
            self.KEY_SUCCESS_GET_STAFF,
            self.KEY_FAILED_PUT_STAFF,
            self.KEY_SUCCESS_PUT_STAFF,
        ]

    def _update(self, event):
        """
        Add operations to the total

        :param event:
        """

        if isinstance(event, StaffOperationEvent):
            goal = event.goal
            total = event.total
            if goal < 0:
                self.indicator_dict[self.KEY_SUCCESS_GET_STAFF] += abs(total)
                self.indicator_dict[self.KEY_FAILED_GET_STAFF] += total - goal
            elif goal > 0:
                self.indicator_dict[self.KEY_SUCCESS_PUT_STAFF] += total
                self.indicator_dict[self.KEY_FAILED_PUT_STAFF] += goal - total


class OccupationKPI(KPI):
    """
    This KPI evaluates the empty and full time and distance
    and the stock relative time/distance
    """

    #: **emptyTime**: time spent empty [seconds]
    KEY_EMPTY_TIME = "emptyTime"
    #: **emptyDistance**: distance travelled empty [meters]
    KEY_EMPTY_DISTANCE = "emptyDistance"
    #: **fullTime**: time spent full [seconds]
    KEY_FULL_TIME = "fullTime"
    #: **fullDistance**: distance travelled full [meters]
    KEY_FULL_DISTANCE = "fullDistance"
    #: **stockTime**: stock relative time (stock*time) [seconds]
    KEY_STOCK_TIME = "stockTime"
    #: **stockDistance**: stock relative distance (stock*distance) [meters]
    KEY_STOCK_DISTANCE = "stockDistance"
    #: **maxStock**: maximum stock
    KEY_MAX_STOCK = "maxStock"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.capacity = None
        self.currentStock = None
        self.previousTime = -1
        self.currentDistance = None

    def _init_keys(self):
        return [
            self.KEY_EMPTY_TIME,
            self.KEY_EMPTY_DISTANCE,
            self.KEY_FULL_TIME,
            self.KEY_FULL_DISTANCE,
            self.KEY_STOCK_TIME,
            self.KEY_STOCK_DISTANCE,
            self.KEY_MAX_STOCK,
        ]

    def reset_for_agent(self, agent):
        super().reset_for_agent(agent)
        self.capacity = None
        self.currentStock = None
        self.previousTime = -1
        self.currentDistance = None

    def get_capacity(self, element):
        """
        Get the capacity of the agent, according to its type.

        :param element:

        :return: agent's capacity
        """
        return self.capacity

    def get_initial_stock(self, element):
        """
        Get the initial stock of the agent, according to its type.

        :param element:

        :return: agent's initial stock
        """
        return self.currentStock

    def end_of_profile_range(self):
        if self.current_profile_index < len(self.profile) - 1:
            self.update_timestamp(self.profile[self.current_profile_index + 1] - 1)
        super().end_of_profile_range()

    def update_timestamp(self, timestamp):
        # compute time spent with last stock
        duration = int(timestamp - self.previousTime)

        # add time to relevant time count
        if self.currentStock is not None:
            if self.currentStock == 0:
                self.indicator_dict[self.KEY_EMPTY_TIME] += duration
                if self.currentDistance is not None:
                    self.indicator_dict[self.KEY_EMPTY_DISTANCE] += self.currentDistance
            elif self.currentStock == self.capacity:
                self.indicator_dict[self.KEY_FULL_TIME] += duration
                if self.currentDistance is not None:
                    self.indicator_dict[self.KEY_FULL_DISTANCE] += self.currentDistance

            # add stock relative time and distance
            self.indicator_dict[self.KEY_STOCK_TIME] += duration * self.currentStock
            if self.currentDistance is not None:
                self.indicator_dict[self.KEY_STOCK_DISTANCE] += (
                    self.currentDistance * self.currentStock
                )

        # set time of last update
        self.previousTime = timestamp

    def add_to_stock(self, value, timestamp):
        """
        Update the full and empty time and distance counts, according to the previous
        stock value, then updates the stock and time.

        :param value: stock change (negative for stock loss)
        :param timestamp: timestamp of the stock change event
        """
        # update indicators for previous period
        self.update_timestamp(timestamp)

        # update current stock
        self.currentStock += value
        if self.currentStock > self.indicator_dict[self.KEY_MAX_STOCK]:
            self.indicator_dict[self.KEY_MAX_STOCK] = self.currentStock

        # reset distance count
        if self.currentDistance is not None:
            self.currentDistance = 0

    def _update(self, event):
        """
        Update the stock and time counts from traced events

        :param event:
        """

        if isinstance(event, InputEvent):
            self.capacity = self.get_capacity(event.element)
            self.currentStock = self.get_initial_stock(event.element)
            self.indicator_dict[self.KEY_MAX_STOCK] = self.currentStock

        if isinstance(event, LeaveSimulationEvent):
            self.update_timestamp(event.timestamp - 1)
            self.currentStock = None


class StationOccupationKPI(OccupationKPI):
    """
    This KPI evaluates the time spent in the empty and full states (of a station),
    and the stock relative time spent in the station
    """

    def _init_keys(self):
        return [self.KEY_EMPTY_TIME, self.KEY_FULL_TIME, self.KEY_STOCK_TIME]

    def new_indicator_dict(self):
        return {
            key: 0
            for key in [
                self.KEY_EMPTY_TIME,
                self.KEY_FULL_TIME,
                self.KEY_STOCK_TIME,
                self.KEY_MAX_STOCK,
            ]
        }

    def get_capacity(self, element):
        return element.capacity

    def get_initial_stock(self, element):
        return element.initial_stock

    def _update(self, event):
        """
        Update the stock and time counts from request events

        :param event:
        """

        super()._update(event)

        if isinstance(event, RequestEvent) and event.request.success:
            request = event.request

            # update time counts and current time
            if request.type == StationRequest.GET_REQUEST:
                self.add_to_stock(-1, request.timestamp)
            elif request.type == StationRequest.PUT_REQUEST:
                self.add_to_stock(1, request.timestamp)

        if isinstance(event, StaffOperationEvent):
            if event.total != 0:
                self.add_to_stock(event.total, event.timestamp)


class VehicleOccupationKPI(OccupationKPI):
    """
    This KPI evaluates the time and distance in the empty and full states (of vehicle),
    and a passenger relative distance and time.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.currentDistance = 0

    def reset_for_agent(self, agent):
        super().reset_for_agent(agent)
        self.currentDistance = 0

    def end_of_profile_range(self):
        super().end_of_profile_range()
        self.currentDistance = 0

    def get_capacity(self, element):
        return element.seats

    def get_initial_stock(self, element):
        # for now, initial stock is always 0 in our simulation
        return 0

    def _update(self, event):
        """
        Update the stock and time/distance counts from
        get/leave vehicle and move events

        :param event:
        """

        super()._update(event)

        if isinstance(event, GetVehicleEvent):
            self.add_to_stock(event.agent.number, event.timestamp)

        elif isinstance(event, LeaveVehicleEvent):
            self.add_to_stock(-event.agent.number, event.timestamp)

        if isinstance(event, MoveEvent):
            self.add_proportioned_indicators(event)

    def evaluate_indicators_on_profile_range(self, event, current_timestamp, duration_on_range):
        if isinstance(event, RouteEvent):
            route_data = event.get_route_data_in_interval(
                current_timestamp, current_timestamp + duration_on_range
            )
            self.currentDistance += sum(route_data["length"])
        else:
            if duration_on_range == 0:
                self.currentDistance += event.distance
            else:
                self.currentDistance += round(event.distance * duration_on_range / event.duration)


class ChargeKPI(KPI):
    """
    This KPI evaluates the trips's boards and un-boards
    """

    PROFILE_COMPATIBILITY = False

    #: **tripId**: gtfs trip id
    KEY_TRIP_ID = "tripId"
    #: **time**: simulation timestamp of board/un-board
    KEY_TIME = "time"
    #: **stopId**: stop id of board/un-board
    KEY_STOP_ID = "stopId"
    #: **boardType**: (+1) for boards, (-1) for un-boards
    KEY_BOARD_TYPE = "boardType"
    #: **value**: numeric value of the charge change
    KEY_VALUE = "value"

    def __init__(self, non_empty_only=True, **kwargs):
        # boolean indicating if only non empty pickups and dropoffs should be traced
        self.non_empty_only = non_empty_only

        super().__init__(**kwargs)

    def new_indicator_dict(self):
        return dict()

    def _init_keys(self):
        return [
            self.KEY_TRIP_ID,
            self.KEY_TIME,
            self.KEY_STOP_ID,
            self.KEY_BOARD_TYPE,
            self.KEY_VALUE,
        ]

    def end_of_events(self):
        pass

    def _update(self, event):
        """
        Add stop information to the list

        :param event:
        """

        if isinstance(event, StopEvent):
            # add a row for each stop type
            if event.dropoffs:
                self.update_stop_information(event)
                self.indicator_dict[self.KEY_TIME] = event.dropoff_time
                self.indicator_dict[self.KEY_BOARD_TYPE] = -1
                self.indicator_dict[self.KEY_VALUE] = len(event.dropoffs)
                self.new_kpi_row()
            if event.pickups:
                self.update_stop_information(event)
                self.indicator_dict[self.KEY_TIME] = event.pickup_time
                self.indicator_dict[self.KEY_BOARD_TYPE] = 1
                self.indicator_dict[self.KEY_VALUE] = len(event.pickups)
                self.new_kpi_row()

            # add a row even if stop is empty if asked
            if not event.pickups and not event.dropoffs and not self.non_empty_only:
                self.update_stop_information(event)
                self.indicator_dict[self.KEY_TIME] = event.timestamp
                self.indicator_dict[self.KEY_BOARD_TYPE] = 0
                self.indicator_dict[self.KEY_VALUE] = 0
                self.new_kpi_row()

    def update_stop_information(self, event):
        """
        Update the indicator with the information common to dropoffs and pickups.

        :param event:
        """

        trip_id = event.trip

        self.indicator_dict[self.KEY_TRIP_ID] = trip_id

        self.indicator_dict[self.KEY_STOP_ID] = get_stop_id_of_event(event)


class PublicTransportChargeKPI(ChargeKPI):
    """
    This KPI evaluates the public transport trips's boards and un-boards
    """

    #: **routeId**: gtfs route id
    KEY_ROUTE_ID = "routeId"
    #: **tripDirection**: gtfs trip direction
    KEY_TRIP_DIRECTION = "tripDirection"

    def __init__(self, **kwargs):
        self.trips = None
        self.routes = None

        super().__init__(**kwargs)

    def _init_keys(self):
        return [self.KEY_ROUTE_ID, self.KEY_TRIP_DIRECTION] + super()._init_keys()

    def _indicators_setup(self, simulation_model):
        self.trips = simulation_model.gtfs.trips
        self.routes = simulation_model.gtfs.routes

    def update_stop_information(self, event):
        super().update_stop_information(event)

        self.indicator_dict[self.KEY_ROUTE_ID] = get_route_id_of_trip(self.trips, event.trip, event)
        self.indicator_dict[self.KEY_TRIP_DIRECTION] = get_direction_of_trip(self.trips, event.trip)


class ServiceKPI(KPI):
    """
    This KPI describes the service time of a vehicle.
    """

    #: **serviceDuration**: vehicle service duration [seconds]
    KEY_SERVICE_DURATION = "serviceDuration"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.last_service_marker = None

    def reset_for_agent(self, agent):
        self.last_service_marker = None
        super().reset_for_agent(agent)

    def end_of_profile_range(self):
        if self.current_profile_index < len(self.profile) - 1:
            self.update_indicator(self.profile[self.current_profile_index + 1] - 1)
        super().end_of_profile_range()

    def update_indicator(self, timestamp):
        if self.last_service_marker is not None:
            self.indicator_dict[self.KEY_SERVICE_DURATION] += timestamp - self.last_service_marker
            self.last_service_marker = timestamp

    def _update(self, event):
        if isinstance(event, ServiceEvent):
            if event.new == SERVICE_UP:
                self.start_service(event.timestamp)
            elif event.new in [SERVICE_PAUSE, SERVICE_END]:
                self.close_service(event.timestamp)
            elif event.new == SERVICE_INIT:
                raise ValueError("Service state shouldn't change to {} status".format(SERVICE_INIT))
            else:
                raise ValueError("Unsupported service status " + event.new)

    def start_service(self, timestamp):
        if self.last_service_marker is not None:
            raise ValueError("Service starts but has not ended")
        else:
            self.last_service_marker = timestamp

    def close_service(self, timestamp):
        if self.last_service_marker is None:
            raise ValueError("Service ends but has not started")
        else:
            self.update_indicator(timestamp)
            self.last_service_marker = None


class TransferKPI(KPI):
    """
    This KPI lists the transfers realised by a user,
    with additional information such as walk distance and duration,
    wait duration, from/to trip/stop
    """

    PROFILE_COMPATIBILITY = False

    #: **walkDistance**: walk distance of transfer [meters]
    KEY_WALK_DIST = "walkDistance"
    #: **walkTime**: walk time of transfer [seconds]
    KEY_WALK_DURATION = "walkDuration"
    #: **waitTime**: wait time of transfer [seconds]
    KEY_WAIT_TIME = "waitTime"
    #: **fromRoute**: origin route of transfer
    KEY_FROM_ROUTE = "fromRoute"
    #: **fromTrip**: origin trip of transfer
    KEY_FROM_TRIP = "fromTrip"
    #: **fromStop**: origin stop point of transfer
    KEY_FROM_STOP = "fromStop"
    #: **toRoute**: destination route of transfer
    KEY_TO_ROUTE = "toRoute"
    #: **toTrip**: destination trip of transfer
    KEY_TO_TRIP = "toTrip"
    #: **toStop**: destination stop of transfer
    KEY_TO_STOP = "toStop"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.trips = None
        self.routes = None

        # transfer variables
        self.current_walk_distance = 0
        self.current_walk_duration = 0
        self.current_wait_time = 0
        self.from_route = None
        self.from_trip = None
        self.from_stop = None
        self.to_route = None
        self.to_trip = None
        self.to_stop = None

    def _indicators_setup(self, simulation_model):
        if simulation_model.gtfs is not None:
            self.trips = simulation_model.gtfs.trips
            self.routes = simulation_model.gtfs.routes

    def _init_keys(self):
        return [
            self.KEY_WALK_DIST,
            self.KEY_WALK_DURATION,
            self.KEY_WAIT_TIME,
            self.KEY_FROM_ROUTE,
            self.KEY_FROM_TRIP,
            self.KEY_FROM_STOP,
            self.KEY_TO_ROUTE,
            self.KEY_TO_TRIP,
            self.KEY_TO_STOP,
        ]

    def end_of_events(self):
        pass

    def new_indicator_dict(self):
        return dict()

    def _update(self, event):
        if isinstance(event, InputEvent):
            self.reset_variables()

        if isinstance(event, WaitEvent):
            self.current_wait_time += event.waiting_time

        elif isinstance(event, MoveEvent) and event.mode == "walk":
            self.current_walk_distance += event.distance
            self.current_walk_duration += event.duration

        elif isinstance(event, RequestEvent):
            self.current_wait_time += sum(event.request.waitSequence)

        elif isinstance(event, StopEvent):
            if isinstance(event.stop, StopPoint):
                stop_id = event.stop.id
            elif isinstance(event.stop, UserStop):
                stop_id = event.stop.stopPoint
            else:
                stop_id = None

            dropoff_agents = [request.agent.id for request in event.dropoffs]
            pickup_agents = [request.agent.id for request in event.pickups]

            if self.agent.id in dropoff_agents:
                self.from_trip = event.trip
                self.from_stop = stop_id

            elif self.agent.id in pickup_agents:
                self.to_trip = event.trip
                self.to_stop = stop_id

                self.write_variables()
                self.reset_variables()

        elif isinstance(event, DestinationReachedEvent):
            self.write_variables()
            self.reset_variables()

    def reset_variables(self):
        self.current_walk_distance = 0
        self.current_walk_duration = 0
        self.current_wait_time = 0
        self.from_trip = None
        self.from_stop = None
        self.to_trip = None
        self.to_stop = None

    def write_variables(self):
        self.indicator_dict[self.KEY_WALK_DIST] = self.current_walk_distance
        self.indicator_dict[self.KEY_WALK_DURATION] = self.current_walk_duration
        self.indicator_dict[self.KEY_WAIT_TIME] = self.current_wait_time
        self.indicator_dict[self.KEY_FROM_ROUTE] = get_route_short_name_of_trip(
            self.trips, self.routes, self.from_trip
        )

        self.indicator_dict[self.KEY_FROM_TRIP] = self.from_trip
        self.indicator_dict[self.KEY_FROM_STOP] = self.from_stop
        self.indicator_dict[self.KEY_TO_ROUTE] = get_route_short_name_of_trip(
            self.trips, self.routes, self.to_trip
        )

        self.indicator_dict[self.KEY_TO_TRIP] = self.to_trip
        self.indicator_dict[self.KEY_TO_STOP] = self.to_stop
        self.new_kpi_row()


# TODO : remove ? useless with TransferKPI
# class JourneyKPI(KPI):
#     """
#     This KPI evaluates the sequence of lines that compose a user journey
#     """
#
#     #: **journeySequence**: sequence of trip ids of the journey
#     KEY_JOURNEY_SEQUENCE = "journeySequence"
#
#     def __init__(self):
#
#         super().__init__()
#
#         self.keys = [self.KEY_JOURNEY_SEQUENCE]
#
#         self.trips_table = None
#         self.routes_table = None
#
#     def setup(self, simulation_model):
#
#         operator = simulation_model.agentPopulation["operators"]["OPR"]
#
#         feed = operator.service_info
#
#         self.trips_table = feed.trips
#         self.routes_table = feed.routes
#
#     def new_indicator_dict(self):
#
#         self.indicator_dict = {self.KEY_JOURNEY_SEQUENCE: ""}
#
#     def update(self, event, agent):
#         """
#         Add a new route for each PickupEvent
#         :param agent:
#         :param event:
#         :return:
#         """
#
#         super().update(event, agent)
#
#         if isinstance(event, PickupEvent):
#
#             route_id = self.trips_table.loc[self.trips_table["trip_id"] == event.trip, "route_id"].iloc[0]
#
#             route_name = self.routes_table.loc[self.routes_table["route_id"] == route_id, "route_short_name"].iloc[0]
#
#             self.indicator_dict[self.KEY_JOURNEY_SEQUENCE] += route_name + "-"
#
#         if isinstance(event, DestinationReachedEvent):
#
#             self.indicator_dict[self.KEY_JOURNEY_SEQUENCE] = self.indicator_dict[self.KEY_JOURNEY_SEQUENCE][:-1]
#


class DestinationReachedKPI(KPI):
    """
    This KPI evaluates the destination reach time
    """

    #: **destinationReachedTime**: time when destination is reached, "NA" otherwise [seconds or NA]
    KEY_DESTINATION_REACHED = "destinationReachedTime"

    def new_indicator_dict(self):
        return {self.KEY_DESTINATION_REACHED: "NA"}

    def _update(self, event):
        """
        Add total wait duration of the request

        :param event:
        """

        if isinstance(event, DestinationReachedEvent):
            self.indicator_dict[self.KEY_DESTINATION_REACHED] = event.timestamp


class LeaveSimulationKPI(KPI):
    """
    This KPI evaluates the cause of the simulation leave.
    """

    #: **leaveSimulation**: code used when leaving the simulation (see model doc)
    KEY_LEAVE_SIMULATION = "leaveSimulation"

    def new_indicator_dict(self):
        return {self.KEY_LEAVE_SIMULATION: None}

    def _update(self, event):
        """
        Add the cause of the LeaveSimulationEvent.

        :param event:
        """

        if isinstance(event, LeaveSimulationEvent):
            self.indicator_dict[self.KEY_LEAVE_SIMULATION] = event.cause


def get_route_id_of_trip(trips, trip_id, event):
    if trips is None:
        return event.serviceVehicle.operator

    trip_table = trips.loc[trips["trip_id"] == trip_id, "route_id"]

    # if trips is not in the gtfs (on-demand trips for instance)
    # try to get operator id
    if trip_table.empty:
        return event.serviceVehicle.operator

    return trip_table.iloc[0]


def get_direction_of_trip(trips, trip_id):
    if trips is None:
        return ""

    trip_table = trips.loc[trips["trip_id"] == trip_id, "direction_id"]

    # ignore trips that are not in the gtfs (on-demand trips for instance)
    if trip_table.empty:
        return ""

    return trip_table.iloc[0]


def get_route_short_name_of_trip(trips, routes, trip_id):
    if trips is None or routes is None or trip_id is None:
        return None

    trip_table = trips.loc[trips["trip_id"] == trip_id, "route_id"]

    # ignore trips that are not in the gtfs (on-demand trips for instance)
    if trip_table.empty:
        return ""

    route_id = trip_table.iloc[0]
    route_short_name = routes.loc[routes["route_id"] == route_id, "route_short_name"].iloc[0]

    return route_short_name


def get_stop_id_of_event(event):
    stop_id = None

    if isinstance(event.stop, StopPoint):
        stop_id = event.stop.id
    elif isinstance(event.stop, UserStop):
        stop_id = event.stop.stopPoint

    if stop_id is None:
        stop_id = ""

    return stop_id



from starling_sim.basemodel.agent.operators.operator import Operator


class StationBasedOperator(Operator):
    """
    Class describing an operator of a station-based shared vehicle service
    """

    SCHEMA = {
        "properties": {
            "stations_dict": {
                "type": "string",
                "title": "Stations population",
                "description": "Population of the operator's station agents",
            },
            "staff_dict": {
                "x-display": "",
                "title": "Staff population",
                "description": "Population of the operator's staff agents",
                "type": "string",
            },
        },
        "required": ["stations_dict"],
        "remove_props": ["stop_points_from"],
    }

    def __init__(self, simulation_model, agent_id, fleet_dict, stations_dict, **kwargs):
        super().__init__(simulation_model, agent_id, fleet_dict, **kwargs)

        # set the station population (agents that store the fleet agents)
        self.stations_dict_name = stations_dict
        self.stations = self.sim.agentPopulation.new_population(stations_dict)

    def init_depot_points(self, depot_points):
        if depot_points is not None and len(depot_points) > 1:
            self.log_message("Multiple depot points are not supported yet. Using first depot.", 30)
            depot_points = [depot_points[0]]

        super().init_depot_points(depot_points)



"""
This module contains tools for the generation of a geojson output,
used for visualisation purposes, after the simulation run
"""

from starling_sim.basemodel.agent.requests import StopPoint
from starling_sim.basemodel.agent.agent import Agent
from starling_sim.basemodel.agent.spatial_agent import SpatialAgent
from starling_sim.basemodel.agent.moving_agent import MovingAgent
from starling_sim.basemodel.agent.operators.operator import Operator
from starling_sim.basemodel.output.feature_factory import *
from starling_sim.basemodel.agent.stations.station import Station
from starling_sim.utils.utils import (
    json_dump,
    new_feature_collection,
    gz_compression,
    gz_decompression,
)
from starling_sim.utils.config import config

from abc import ABC

# list of supported icons for the geojson format
ICON_LIST = [
    "user",
    "user-01",
    "user-02",
    "user-03",
    "user-04",
    "user-05",
    "user-06",
    "user-07",
    "user-08",
    "user-09",
    "user-10",
    "stop_point",
    "station",
    "car",
    "bike",
    "bus",
    "subway",
    "tram",
    "train",
    "odt",
    "truck",
]

# default version of the geojson output format
CURRENT_GEOJSON_VERSION = 1


# Abstract Base Class for the geojson output generation


class GeojsonOutput(ABC):
    """
    Manage the generation of a geojson output, mainly for visualisation purposes.
    """

    #: Version of the geojson output generator
    VERSION = "X.X.X"

    def __init__(self):
        """
        Create a new object for the generation of a geojson file.

        Most attributes are initialised within the setup method.
        """

        # simulation information

        # simulation model from which information is extracted
        self.sim = None

        # dict of the simulation topologies
        self.graphs = None

        # construction of the output structure as a FeatureCollection

        # dict giving the information factories for each agent type
        self.information_factories = dict()

        # list of features of the final output
        self.features = None

        # current element converted as a feature
        self.current_element = None

        # feature of the current agent
        self.current_feature = None

        # output file

        # output folder
        self.folder = None

        # output filename
        self.filename = None

    def setup(self, simulation_model, filename, folder):
        """
        Set the values of the simulation model, topologies and outfile

        This setup method is called during simulation setup

        :param simulation_model:
        :param filename:
        :param folder:
        """

        self.sim = simulation_model
        self.graphs = simulation_model.environment.topologies
        self.features = []
        self.filename = filename
        self.folder = folder

        for factories in self.information_factories.values():
            for factory in factories:
                factory.setup(simulation_model)

    def set_information_factories(self, information_factories):
        self.information_factories = information_factories

    def add_population_features(self, population=None):
        """
        Build and add features for the given population, with additional information
        provided by the information factories.

        These are added to the features attributes, used to build the final FeatureCollection.

        If population is

        :param population: population key, population list or None.
        """

        if isinstance(population, str) and population in self.sim.agentPopulation:
            population_list = self.sim.agentPopulation[population].values()
        elif isinstance(population, list):
            population_list = population
        elif population is None:
            population_list = self.sim.agentPopulation.get_total_population()
        else:
            raise TypeError("Unsupported type for population : {}".format(population.__class__))

        for element in population_list:
            self.add_element_feature(element)

    def add_element_feature(self, element):
        self.init_element_feature(element)

        self.add_factories_information()

        if self.current_feature is not None:
            self.features.append(self.current_feature)

    def init_element_feature(self, element):
        pass

    def add_factories_information(self):
        pass

    def generate_geojson(self):
        """
        Write the geojson feature collection in a file.
        """

        # generate a feature collection from the geojson features
        feature_collection = new_feature_collection(self.features)

        # add a version field
        feature_collection["version"] = self.VERSION

        # build file path
        path = self.folder + self.filename

        # check bz2 extension
        if path.endswith(".gz"):
            to_gz = True
            path = path[:-3]
        else:
            to_gz = False

        # write the geojson file
        json_dump(feature_collection, path)

        # compress to bz2 if necessary
        mimetype = "application/json"
        if to_gz:
            path = gz_compression(path)
            mimetype = "application/gzip"

        # signal new file to output factory
        self.sim.outputFactory.new_output_file(
            path, mimetype, compressed_mimetype="application/json", content="visualisation"
        )


# Classes for the generation of geojson output


class GeojsonOutput1(GeojsonOutput):
    VERSION = "1.0"

    def init_element_feature(self, element):
        """
        Initialise a geojson feature for the element regarding its class.

        The current_feature attribute is set with a geojson feature or None.

        :param element: simulation element
        """

        self.current_element = element
        self.current_feature = None

        if isinstance(element, MovingAgent):
            self.current_feature = create_line_string_feature(self, element)

        elif isinstance(element, StopPoint):
            self.current_feature = create_point_feature(self, element, icon_type="stop_point")

        elif isinstance(element, SpatialAgent):
            self.current_feature = create_point_feature(self, element)

        elif isinstance(element, Operator):
            if len(element.servicePoints.values()) != 0:
                self.add_population_features(list(element.servicePoints.values()))

            self.current_element = element
            if element.serviceZone is not None:
                self.current_feature = create_multi_polygon_feature(self, element, icon_type="")

    def add_factories_information(self):
        if not isinstance(self.current_element, Agent):
            return

        if self.current_element.type not in self.information_factories:
            return

        information_dict = dict()
        factories = self.information_factories[self.current_element.type]

        for event in self.current_element.trace.eventList:
            for factory in factories:
                factory.update(event, self.current_element)

        for factory in factories:
            information_dict[factory.key] = factory.get_dict()

        self.current_feature["properties"]["information"] = information_dict


class GeojsonOutput0(GeojsonOutput):
    VERSION = "0.1"

    def init_element_feature(self, element):
        self.current_element = element
        self.current_feature = None

        if isinstance(element, Operator):
            if len(element.servicePoints.values()) != 0:
                self.add_population_features(list(element.servicePoints.values()))

            self.current_element = element
            if element.serviceZone is not None:
                self.current_feature = create_multi_polygon_feature(self, element, icon_type="")

        elif isinstance(element, StopPoint):
            self.current_feature = create_point_feature(self, element, icon_type="stop_point")
            self.current_feature["properties"]["position"] = [
                self.current_feature["geometry"]["coordinates"]
            ] * 2
            self.current_feature["properties"]["duration"] = [0, 99999]

        elif isinstance(element, Station):
            self.current_feature = create_point_feature(self, element, icon_type="station")
            self.current_feature["properties"]["position"] = [
                self.current_feature["geometry"]["coordinates"]
            ] * 2
            self.current_feature["properties"]["duration"] = [0, 99999]

        elif isinstance(element, SpatialAgent):
            self.current_feature = create_point_feature(self, element)
            localisations, timestamps = get_element_line_string(self, element)
            self.current_feature["geometry"]["coordinates"] = localisations[0]
            self.current_feature["properties"]["position"] = localisations
            self.current_feature["properties"]["duration"] = timestamps


# function for the creation of a geojson output instance of the correct version


def new_geojson_output():
    """
    Create a new GeojsonOutput instance.

    Use the version provided in the 'geojson_version' field of parameters,
    and the default version if not provided.

    :return: GeojsonOutput instance
    """

    # get the geojson version
    geojson_version = config["geojson_version"]
    if geojson_version is None:
        geojson_version = CURRENT_GEOJSON_VERSION

    # create the geojson output
    if geojson_version == 0:
        geojson_output = GeojsonOutput0()
    elif geojson_version == 1:
        geojson_output = GeojsonOutput1()
    else:
        logging.warning(
            "Unknown geojson format version {}. Give only the main version number.".format(
                geojson_version
            )
        )
        geojson_output = None

    return geojson_output



"""
This module contains the classes dedicated to trace agent activities
during the simulation. Events describes different kind of traced
activities and are generated during the simulation.

/!\\ It is should be mentioned that the event objects store in the trace
point to objects that may be modified by the end of the simulation.
For instance, given a GenerationEvent, display the 'position' attribute of the
generated agent at the end of the simulation will output its final
position in the simulation, and not its origin.
"""

from starling_sim.utils.simulation_logging import TRACED_LOGGER, AGENT_LEVEL
from starling_sim.basemodel.trace.events import LeaveSimulationEvent
from starling_sim.utils.constants import END_OF_SIM_LEAVE


class Trace:
    """
    This class represents the trace of a simulation element

    It contains events that describe their states and actions
    """

    def __init__(self):
        """
        Start with empty event list
        """
        self.eventList = []

    def add_event(self, event):
        """
        :param event: Event object, describing the traced event
        :return:
        """
        self.eventList.append(event)


class Traced:
    """
    This class represents a traced element of the simulation

    Other classes may extend it to be able to trace their actions
    """

    def __init__(self, element_id):
        """
        Construction of a new traced element
        The trace must have access to the simulation model
        :param: element_id: traced element id
        """

        self.sim = None
        self.id = element_id
        self.trace = Trace()

    def __str__(self):
        """
        Give a string display to the simulation element
        """

        return "[id={}]".format(self.id)

    def __repr__(self):
        """
        Give a string display to the simulation element
        """

        return self.__str__()

    def trace_event(self, event):
        """
        Adds a Event to the trace, with the current timestamp
        :param event: Traced event
        :return:
        """
        self.trace.add_event(event)

    def log_message(self, message, lvl=AGENT_LEVEL):
        """
        Logs the message in the logger using the class data

        :param message: Message displayed in agent log
        :param lvl: level value, default is AGENT_LEVEL
        """

        extra_params = {"id": self.id}

        if self.sim is not None:
            extra_params["timestamp"] = self.sim.scheduler.now()
        else:
            extra_params["timestamp"] = "--"

        TRACED_LOGGER.log(lvl, message, extra=extra_params)


def trace_simulation_end(simulation_model):
    """
    Adds a LeaveSimulationEvent with cause end of simulation to
    all agents that didn't leave during the simulation.
    """

    for agent in simulation_model.agentPopulation.get_total_population():
        if not isinstance(agent.trace.eventList[-1], LeaveSimulationEvent):
            agent.trace_event(
                LeaveSimulationEvent(simulation_model.scheduler.now(), agent, END_OF_SIM_LEAVE)
            )
